import unittest
import numpy as np
import raynest.model

class GaussianMixtureModel(raynest.model.Model):
    """
    A simple gaussian model with parameters mean and sigma
    Shows example of using your own data
    """
    def __init__(self):
        self.names=['mean1','sigma1','mean2','sigma2','weight']
        self.bounds=[[-3,3],[0.01,1],[-3,3],[0.01,1],[0.0,1.0]]
        data = []
        for _ in range(10000):
            if np.random.uniform(0.0,1.0) < 0.1:
                data.append(np.random.normal(0.5,0.5))
            else:
                data.append(np.random.normal(-1.5,0.03))

        self.data = np.array(data)

    def log_likelihood(self,x):
        w = x['weight']
        logL1 = np.log(w)-np.log(x['sigma1'])-0.5*((self.data-x['mean1'])/x['sigma1'])**2
        logL2 = np.log(1.0-w)-np.log(x['sigma2'])-0.5*((self.data-x['mean2'])/x['sigma2'])**2
        logL = np.logaddexp(logL1, logL2).sum()
        return logL

    def log_prior(self,p):
        if not self.in_bounds(p): return -np.inf
        return -np.log(p['sigma1'])-np.log(p['sigma2'])

    def force(self,x):
        f = np.zeros(1, dtype = {'names':x.names, 'formats':['f8' for _ in x.names]})
        f['sigma1'] = 1.0
        f['sigma2'] = 1.0
        return f

class GaussianMixtureTestCase(unittest.TestCase):
    """
    Test the gaussian model
    """
    def setUp(self):
        self.work=raynest.raynest(GaussianMixtureModel(),verbose=2,nensemble=0,nlive=1024,maxmcmc=1000,nslice=6)

    def test_run(self):
        self.work.run()



def test_all():
    unittest.main(verbosity=2)

if __name__=='__main__':
    unittest.main(verbosity=2)



from __future__ import division
import sys
import os
import logging
import numpy as np
from math import log
from collections import deque

from . import parameter
from .proposal import DefaultProposalCycle, EnsembleProposal
from . import proposal
import numpy.lib.recfunctions as rfn
import array
from .nest2pos import acl, autocorrelation
import ray
import gc

import pickle
__checkpoint_flag = False


class Sampler(object):
    """
    Sampler class.
    ---------

    Initialisation arguments:

    args:
    model: :obj:`raynest.Model` user defined model to sample

    maxmcmc:
        :int: maximum number of mcmc steps to be used in the :obj:`cnest.sampler.Sampler`

    ----------
    kwargs:

    verbose:
        :int: display debug information on screen
        Default: 0

    proposal:
        :obj:`raynest.proposals.Proposal` to use
        Defaults: :obj:`raynest.proposals.DefaultProposalCycle`)
    """

    def __init__(self,
                 model,
                 maxmcmc,
                 rng          = None,
                 verbose      = 0,
                 proposal     = None):

        self.counter        = 0
        self.model          = model
        if rng == None:
            self.rng            = np.random.default_rng()
        else:
            self.rng            = rng
        self.initial_mcmc   = maxmcmc//10
        self.maxmcmc        = maxmcmc
        self.tau            = None
        self.logger         = logging.getLogger('raynest')

        if proposal is None:
            self.proposal = DefaultProposalCycle(self.rng)
        else:
            self.proposal = proposal(self.rng, model = self.model)
        
        self.Nmcmc              = self.initial_mcmc
        self.Nmcmc_exact        = float(self.initial_mcmc)

        self.verbose            = verbose
        self.acceptance         = 0.0
        self.sub_acceptance     = 0.0
        self.mcmc_accepted      = 0
        self.mcmc_counter       = 0
        self.initialised        = False
        self.max_storage        = 32768
        self.samples            = deque(maxlen = None if self.verbose >=3 else self.max_storage) # the list of samples from the mcmc chain

    def estimate_nmcmc_on_the_fly(self, safety=5, tau=None):
        """
        Estimate autocorrelation length of chain using acceptance fraction
        ACL = (2/acc) - 1
        multiplied by a safety margin of 5
        Uses moving average with decay time tau iterations
        (default: :int:`self.maxmcmc`)

        Taken from http://github.com/farr/EnsembleNest.jl
        """
        # if no auto correlation time is provided, make the safest assumption
        # that maxmcmc is safety times the auto correlation time
        if tau is None: tau  = self.maxmcmc

        # if we accepted no points, increase the number of steps
        if self.sub_acceptance == 0.0:
            self.Nmcmc_exact = (1.0 + 1.0/tau)*self.Nmcmc_exact
        else:
            self.Nmcmc_exact = (1.0 - 1.0/tau)*self.Nmcmc_exact + (safety/tau)*(2.0/self.sub_acceptance - 1.0)

        self.Nmcmc_exact = float(min(self.Nmcmc_exact,self.maxmcmc))
        self.Nmcmc = max(safety,int(self.Nmcmc_exact))

        return self.Nmcmc

    def estimate_acl(self, safety=1):
        """
        Estimate autocorrelation length of the chain
        
        safety determines the number of ACL the chains are going to be run for
        """
        # first of all, build a numpy array out of
        # the stored samples
        ACL = []
        s   = list(self.samples)
        samples = np.array([x.values for x in s[-self.max_storage:]])
        # compute the ACL on the max_storage set of samples, choosing a window length of 5*tau
        ACL = [acl(samples[:,i], c=5) for i in range(samples.shape[1])]
        
        if self.verbose >= 3:
            for i in range(len(self.model.names)):
                self.logger.info("Sampler {0} -- ACL({1})  = {2}".format(os.getpid(),self.model.names[i],ACL[i]))
        
        # take the longest ACL as the auto correlation time of the chain
        self.tau = np.max(ACL)
        # set the number of steps to be safety times the ACL
        self.Nmcmc = int(safety*self.tau)
        return self.Nmcmc

    def produce_sample(self, old, logLmin):
        """
        main loop that takes the worst :obj:`raynest.parameter.LivePoint` and
        evolves it. Proposed sample is then sent back
        to :obj:`raynest.NestedSampler`.
        """
        (Nmcmc, outParam) = self.return_sample(old, logLmin)
        # Send the sample to the Nested Sampler

        self.counter += 1
        if self.counter%10 == 0:
            self.estimate_acl()
        
        return self.acceptance, self.sub_acceptance, Nmcmc, outParam

    def set_ensemble(self, ensemble_ref):
        if self.verbose > 3:
            self.logger.info("Sampler {0} -- setting ensemble".format(os.getpid()))
        
        ensemble = ray.get(ensemble_ref)[0]
        for p in self.proposal.proposals:
            if isinstance(p, EnsembleProposal):
                p.set_ensemble(ensemble)
        del ensemble
        
        return 0

@ray.remote
class MetropolisHastingsSampler(Sampler):
    """
    metropolis-hastings acceptance rule
    for :obj:`raynest.proposal.EnsembleProposal`
    """
    def return_sample(self, oldparam, logLmin):

        sub_counter = 0
        sub_accepted = 0
        logp_old = self.model.log_prior(oldparam)

        while True:

            sub_counter += 1
            newparam = self.proposal.get_sample(oldparam.copy())
            newparam.logP = self.model.log_prior(newparam)

            if newparam.logP-logp_old + self.proposal.log_J > np.log(self.rng.random()):

                newparam.logL = self.model.log_likelihood(newparam)

                if newparam.logL > logLmin:

                    oldparam = newparam.copy()
                    logp_old = newparam.logP
                    sub_accepted+=1

            # append the sample to the array of samples
            self.samples.append(oldparam)

            if (sub_counter >= self.Nmcmc and sub_accepted > 0 ) or sub_counter >= self.maxmcmc:
                break

        self.sub_acceptance = float(sub_accepted)/float(sub_counter)
        self.mcmc_accepted += sub_accepted
        self.mcmc_counter += sub_counter
        self.acceptance    = float(self.mcmc_accepted)/float(self.mcmc_counter)
        # return the new sample
        return (sub_counter, oldparam)


@ray.remote
class HamiltonianMonteCarloSampler(Sampler):
    """
    HamiltonianMonteCarlo acceptance rule
    for :obj:`raynest.proposal.HamiltonianProposal`
    WARNING: this is not working properly quite yet
    """
    initialised = False
    density     = None
    
    def set_ensemble(self, ensemble_ref):

        ensemble = ray.get(ensemble_ref)[0]
        # compute the covariance
        ensemble.update_mean_covariance()
        covariance = ensemble.get_covariance()
            
        self.density  = ADPGMM(self.model.bounds, probit = False)
        for e in ensemble._list:
            self.density.add_new_point(e.values)
            
        self.proposal.set_ensemble((self.density, covariance))
        return 0
    
    def return_sample(self, oldparam, logLmin):
        
        sub_accepted    = 0
        sub_counter     = 0

        while True:

            sub_counter += 1
            newparam     = self.proposal.get_sample(oldparam.copy(), logLmin = logLmin)

            if self.proposal.log_J > np.log(self.rng.random()):

                if newparam.logL > logLmin:
                    oldparam        = newparam.copy()
                    sub_accepted   += 1

            # append the sample to the array of samples
            self.samples.append(oldparam)
        
            if sub_counter >= self.Nmcmc and sub_accepted > 0:
                break

#            if sub_counter >= self.maxmcmc:
#                break

        self.sub_acceptance = float(sub_accepted)/float(sub_counter)
        self.mcmc_accepted += sub_accepted
        self.mcmc_counter  += sub_counter
        self.acceptance     = float(self.mcmc_accepted)/float(self.mcmc_counter)

        self.density.add_new_point(oldparam.values)
        
        return (sub_counter, oldparam)
         
@ray.remote
class SliceSampler(Sampler):
    """
    The Ensemble Slice sampler from Karamanis & Beutler
    https://arxiv.org/pdf/2002.06212v1.pdf
    """
    mu             = 1.0
    max_steps_out  = 1000 # maximum stepping out steps allowed
    max_slices     = 1000 # maximum number of slices allowed
    
    def adapt_length_scale(self):
        """
        adapts the length scale of the expansion/contraction
        following the rule in (Robbins and Monro, 1951) of Tibbits et al. (2014)
        """
        Ne = max(1,self.Ne)
        Nc = max(1,self.Nc)
        ratio = Ne/(Ne+Nc)
        self.mu *= 2*ratio

    def reset_boundaries(self):
        """
        resets the boundaries and counts
        for the slicing
        """
        self.L  = - self.rng.random()
        self.R  = self.L + 1.0
        self.Ne = 0.0
        self.Nc = 0.0

    def increase_left_boundary(self):
        """
        increase the left boundary and counts
        by one unit
        """
        self.L  = self.L - 1.0
        self.Ne = self.Ne + 1

    def increase_right_boundary(self):
        """
        increase the right boundary and counts
        by one unit
        """
        self.R  = self.R + 1.0
        self.Ne = self.Ne + 1

    def return_sample(self, oldparam, logLmin):

        sub_accepted    = 0
        sub_counter     = 0

        while True:
            # Set Initial Interval Boundaries
            self.reset_boundaries()
            sub_counter += 1

            direction_vector = self.proposal.get_direction(mu = self.mu)
            if not(isinstance(direction_vector,parameter.LivePoint)):
                direction_vector = parameter.LivePoint(oldparam.names,d=direction_vector)

            Y = logLmin
            Yp = oldparam.logP-self.rng.exponential()
            J = np.floor(self.max_steps_out*self.rng.random())
            K = (self.max_steps_out-1)-J
            # keep on expanding until we get outside the logL boundary from the left
            # or the prior bound, whichever comes first
            while J > 0:

                parameter_left = direction_vector * self.L + oldparam

                if self.model.in_bounds(parameter_left):
                    if Yp > self.model.log_prior(parameter_left):
                        break
                    else:
                        self.increase_left_boundary()
                        J -= 1
                # if we get out of bounds, break out
                else:
                    break

            # keep on expanding until we get outside the logL boundary from the right
            # or the prior bound, whichever comes first
            while K > 0:

                parameter_right = direction_vector * self.R + oldparam

                if self.model.in_bounds(parameter_right):

                    if Yp > self.model.log_prior(parameter_right):
                        break
                    else:
                        self.increase_right_boundary()
                        K -= 1
                # if we get out of bounds, break out
                else:
                    break

            # slice sample the likelihood-bound prior
            slice = 0
            while slice < self.max_slices:
                # generate a new point between the boundaries we identified
                Xprime        = self.rng.uniform(self.L,self.R)
                newparam      = direction_vector * Xprime + oldparam
                newparam.logP = self.model.log_prior(newparam)

                if newparam.logP > Yp:
                    # compute the new value of logL
                    newparam.logL = self.model.log_likelihood(newparam)
                    if newparam.logL > Y:
                        oldparam     = newparam.copy()
                        sub_accepted += 1
                        break
                # adapt the intervals shrinking them
                if Xprime < 0.0:
                    self.L = Xprime
                    self.Nc = self.Nc + 1
                elif Xprime > 0.0:
                    self.R = Xprime
                    self.Nc = self.Nc + 1

                slice += 1

            if sub_counter >= self.Nmcmc and sub_accepted > 0:
                break

            if sub_counter >= self.maxmcmc:
                break

            self.samples.append(oldparam)
        self.sub_acceptance = float(sub_accepted)/float(sub_counter)
        self.mcmc_accepted += sub_accepted
        self.mcmc_counter  += sub_counter
        self.acceptance     = float(self.mcmc_accepted)/float(self.mcmc_counter)

        if logLmin == -np.inf:
            self.adapt_length_scale()

        return (sub_counter, oldparam)

@ray.remote
class SamplersCycle(Sampler):
    """
    A sampler that cycles through a list of
    samplers.

    Initialisation arguments:

    samplers : A list of samplers

    Optional arguments:
    cyclelength : length of the sampler cycle. Default: 100

    """
    idx = 0 # index in the cycle
    N   = 0   # number of samplers in the cycle
    def __init__(self,samplers,*args,**kwargs):
        self.samplers = samplers
        self.N        = len(self.samplers)

    def produce_sample(self, old, logLmin, **kwargs):
        # Call the current sampler and increment the index
        self.idx = (self.idx + 1) % self.N
        p = self.samplers[self.idx]
        acceptance, sub_acceptance, Nmcmc, new = p.produce_sample(old, logLmin, **kwargs)
        return acceptance, sub_acceptance, Nmcmc, new

    def set_ensemble(self, ensemble):
        """
        Updates the ensemble statistics
        by calling it on each :obj:`EnsembleProposal`
        """
        for s in self.samplers:
            for p in s.proposal.proposals:
                if isinstance(p, EnsembleProposal):
                    p.set_ensemble(ensemble)



import unittest
import numpy as np
import raynest.model

class RingModel(raynest.model.Model):
    """
    A circular ring in parameter space
    logZ ~ -2.31836
    """
    names=['x','y']
    bounds=[[-2,2],[-2,2]]
    data = None
    analytic_log_Z = -2.31836
    @staticmethod
    def log_likelihood(x):
        return log_ring(x['x'],x['y'])
    def force(self, p):
        f = np.zeros(1, dtype = {'names':p.names, 'formats':['f8' for _ in p.names]})
        return f

def log_ring(x, y, radius=1.0, thickness=0.1):
    r = np.sqrt(x**2+y**2)
    return -0.5*(radius-r)**2/thickness**2

class RingTestCase(unittest.TestCase):
    """
    Test the gaussian model
    """
    def setUp(self):
        self.model=RingModel()
        self.work=raynest.raynest(self.model,verbose=1,nslice=4,nlive=1000,maxmcmc=1000)
        self.work.run()


    def test_evidence(self):
        # 2 sigma tolerance
        logZ = self.work.logZ
        H    = self.work.information
        tolerance = 2.0*self.work.logZ_error
        print('2-sigma statistic error in logZ: {0:0.3f}'.format(tolerance))
        print('Analytic logZ {0}'.format(self.model.analytic_log_Z))
        print('Estimated logZ {0}'.format(logZ))
        self.assertTrue(np.abs(logZ - RingModel.analytic_log_Z)<tolerance, 'Incorrect evidence for normalised distribution: {0:.3f} +/ {2:.3f} instead of {1:.3f}'.format(logZ,RingModel.analytic_log_Z,tolerance ))

def test_all():
    unittest.main(verbosity=2)

if __name__=='__main__':
    unittest.main(verbosity=0)




try:
    from figaro.mixture import DPGMM
    from figaro.likelihood import log_norm
except:
    print("FIGARO not available! Hamiltonian Monte Carlo sampling not supported!")
    pass

import numpy as np
from tqdm import tqdm
from scipy.stats import multivariate_normal as mn

class ADPGMM(DPGMM):
    def __init__(self, *args, **kwargs):
        super(ADPGMM, self).__init__(*args, **kwargs)
        
    def _pdf_array(self, x):
        """
        Evaluate mixture at point(s) x in probit space
        
        Arguments:
            :np.ndarray x: point(s) to evaluate the mixture at (in probit space)
        
        Returns:
            :np.ndarray: comp.pdf(x)
        """
        return np.array([w*np.exp(log_norm(x[0], comp.mu, comp.sigma)) for comp, w in zip(self.mixture, self.w)])

    def _log_gradient(self, x):
        """
        Returns the log gradient of the mixture
        
        Arguments:
            :np.ndarray x: point to evaluate the gradient at
        
        Returns:
            :np.ndarray: log_gradient
        """
        p = self._pdf_array(x)
        B = np.array([-np.dot(np.linalg.inv(comp.sigma),(x - comp.mu)) for comp in self.mixture])

        return np.average(B, weights = p, axis = 0)

if __name__ == "__main__":
    dim = 1
    N = 10000
    data = np.random.normal(0.0,1.0, size = (N, dim))

    mix = ADPGMM([[-10, 10]], probit = False)
    for s in tqdm(data):
        mix.add_new_point(s)

    mix.build_mixture()
    print(mix._log_gradient([0.0]))

    import matplotlib.pyplot as plt

    x = np.linspace(-10,10,101)
    G = [mix._log_gradient(xi) for xi in x]
    f = plt.figure()
    ax = f.add_subplot(121)
    ax.plot(x,mix.pdf(x))
    ax.hist(data,density=True,bins=100)
    ax = f.add_subplot(122)
    ax.plot(x,G,color='r')
    ax.plot(x,np.gradient(mix.logpdf(x),np.diff(x)[0]),color='k')
    plt.show()








import logging
import numpy as np
from numpy import logaddexp, vstack
from numpy.random import uniform
from functools import reduce
from scipy.signal import correlate
# if not logging.Logger.manager.loggerDict:
#     LOGGER = logging.getLogger('raynest.nest2pos')
# else:
#     LOGGER = logging.getLogger('raynest.raynest.raynest')

# Note - logger should take the name of the module. It inherits from the base
# raynest logger.
LOGGER = logging.getLogger('raynest.nest2pos')

def logsubexp(x,y):
    """
    Helper function to compute the exponential
    of a difference between two numbers

    ----------
    Parameter:
        x: :float:
        y: :float:
    ----------
    Return
        z: :float: x + np.log1p(-np.exp(y-x))
    """
    assert np.all(x >= y), 'cannot take log of negative number {0!s} - {1!s}'.format(str(x), str(y))
    return x + np.log1p(-np.exp(y-x))

def log_integrate_log_trap(log_func,log_support):
    """
    Trapezoidal integration of given log(func)
    Returns log of the integral
    """

    log_func_sum = logaddexp(log_func[:-1], log_func[1:]) - np.log(2)
    log_dxs = logsubexp(log_support[:-1], log_support[1:])

    return logaddexp.reduce(log_func_sum + log_dxs)


def compute_weights(data, Nlive):
    """Returns log_ev, log_wts for the log-likelihood samples in data,
    assumed to be a result of nested sampling with Nlive live points."""

    start_data=np.concatenate(([float('-inf')], data[:-Nlive]))
    end_data=data[-Nlive:]

    log_wts=np.zeros(data.shape[0])

    log_vols_start=np.cumsum(np.ones(len(start_data)+1)*np.log1p(-1./Nlive))-np.log1p(-1./Nlive)
    log_vols_end=np.zeros(len(end_data))
    log_vols_end[-1]=np.NINF
    log_vols_end[0]=log_vols_start[-1]+np.log1p(-1.0/Nlive)
    for i in range(len(end_data)-1):
        log_vols_end[i+1]=log_vols_end[i]+np.log1p(-1.0/(Nlive-i))

    log_likes = np.concatenate((start_data,end_data,[end_data[-1]]))

    log_vols=np.concatenate((log_vols_start,log_vols_end))
    log_ev = log_integrate_log_trap(log_likes, log_vols)

    log_dXs = logsubexp(log_vols[:-1], log_vols[1:])
    log_wts = log_likes[1:-1] + log_dXs[:-1]

    log_wts -= log_ev

    return log_ev, log_wts

def draw_posterior(data, log_wts, verbose=False, rng=None):
    """Draw points from the given data (of shape (Nsamples, Ndim))
    with associated log(weight) (of shape (Nsamples,)). Draws uniquely so
    there are no repeated samples"""
    if rng is None:
        rng = uniform

    maxWt=max(log_wts)
    normalised_wts=log_wts-maxWt
    selection=[n > np.log(rng()) for n in normalised_wts]
    idx=list(filter(lambda i: selection[i], range(len(selection))))
    return data[idx]

def draw_posterior_many(datas, Nlives, verbose=False, rng=None):
    """Draw samples from the posteriors represented by the
    (Nruns, Nsamples, Nparams)-shaped array datas, each sampled with
    the corresponding Nlive number of live points. Will draw without repetition,
    and weight according to the evidence in each input run"""
    
    if rng is None:
        generator = uniform
    else:
        generator = rng.random
    
    # list of log_evidences, log_weights
    
    log_evs,log_wts=zip(*[compute_weights(data['logL'],Nlive) for data,Nlive in zip(datas, Nlives)])
    if verbose:
        LOGGER.critical('Computed log_evidences: {0!s}'.format((str(log_evs))))

    log_total_evidence=reduce(logaddexp, log_evs)
    log_max_evidence=max(log_evs)
    #print 'evidences: %s'%(str(log_evs))
    fracs=[np.exp(log_ev-log_max_evidence) for log_ev in log_evs]
    if verbose:
        LOGGER.critical('Relative weights of input files: {0!s}'.format((str(fracs))))
    Ns=[fracs[i]/len(datas[i]) for i in range(len(fracs))]
    Ntot=max(Ns)
    fracs=[n/Ntot for n in Ns]
    if verbose:
        LOGGER.critical('Relative weights of input files taking into account their length: {0!s}'.format((str(fracs))))

    posts=[draw_posterior(data,logwt, rng = generator) for (data,logwt,logZ) in zip(datas,log_wts,log_evs)]
    if verbose:
        LOGGER.critical('Number of input samples: {0!s}'.format((str([len(x) for x in log_wts]))))
        LOGGER.critical('Expected number of samples from each input file {0!s}'.format((str([int(f*len(p)) for f,p in zip(fracs,posts)]))))
    bigpos=[]
    for post,frac in zip(posts,fracs):
      mask = generator(size=len(post))<frac
      bigpos.append(post[mask])
    result = np.concatenate([bigpos[i] for i in range(len(bigpos))], axis=None)
    if verbose:
        LOGGER.critical('Samples produced: {0:d}'.format(result.shape[0]))
    return result, log_total_evidence-np.log(len(datas))

def draw_N_posterior(data,log_wts, N, verbose=False):
    """
    Draw N samples from the input data, weighted by log_wt.
    For large N there may be repeated samples
    """
    if(N==0): return []
    log_cumsums=np.zeros(log_wts.shape[0]+1)
    log_cumsums[0]=-float('inf')
    for i,log_wt in enumerate(log_wts):
        log_cumsums[i+1]=logaddexp(log_cumsums[i], log_wt)

    us=np.log(uniform(size=N))
    idxs=np.digitize(us, log_cumsums)

    return data[idxs-1]

def draw_N_posterior_many(datas, Nlives, Npost, verbose=False):
    """
    Draw Npost samples from the posteriors represented by the
    (Nruns, Nsamples, Nparams)-shaped array datas, each sampled with
    the corresponding number of live points Nlive. The returned number
    of samples may not be exactly Npost due to rounding
    """
    # get log_evidences, log_weights.
    log_evs,log_wts=zip(*[compute_weights(data['logL'],Nlive) for data,Nlive in zip(datas, Nlives)])

    log_total_evidence=reduce(logaddexp, log_evs)
    Ns=[int(round(Npost*np.exp(log_ev-log_total_evidence))) for log_ev in log_evs]
    posts=[draw_N_posterior(data,logwt,N) for (data,logwt,N) in zip(datas,log_wts,Ns)]
    return vstack(posts).flatten()

def resample_mcmc_chain(chain, verbose=False, burnin=False):
    """
    Draw samples from the mcmc chains posteriors by redrawing
    each one of them against the Metropolis-Hastings rule
    """
    LOGGER.critical('Number of input samples: {0!s}'.format(chain.shape[0]))
    if burnin: chain = chain[chain.shape[0]/2-1:]
    # thin according to the autocorrelation length
    ACL = []
    for n in chain.dtype.names:
        if n != 'logL' and n != 'logPrior':
            ACL.append(acl(chain[n]))
    ACL = int(np.round(np.max(ACL)))

    LOGGER.critical('Measured autocorrelation length {0!s}'.format(str(ACL)))
    # thin and shuffle the chain
    chain = chain[::ACL]
    np.random.shuffle(chain)
    # compute the normalised log posterior density
    logpost = chain['logL']+chain['logPrior']
    # resample using a Metropolis-Hastings rule
    output =  [chain[0]]
    for i in range(1,chain.shape[0]):
        if logpost[i] - logpost[i-1] > np.log(uniform()):
            output.append(chain[i])

    output = np.array(output)
    LOGGER.critical('Returned number of samples {0!s}'.format(str(output.shape[0])))

    return output

def next_pow_two(n):
    """
    find the next power of 2 given n
    """
    i = 1
    while i < n:
        i = i << 1
    return i

def autocorrelation(x, norm=True):
    """
    compute the autocorrelation function of an array
    """
    x = np.atleast_1d(x)
    if len(x.shape) != 1:
        raise ValueError("invalid dimensions for 1D autocorrelation function")
    n = next_pow_two(len(x))

    # Compute the FFT and then (from that) the auto-correlation function
    f = np.fft.fft(x - np.mean(x), n=2 * n)
    acf = np.fft.ifft(f * np.conjugate(f))[: len(x)].real
    acf /= 4 * n

    # Optionally normalize
    if norm:
        acf /= acf[0]

    return acf

def auto_window(taus, c):
    """
    Automated windowing procedure following Sokal (1989)
    """
    m = np.arange(len(taus)) < c * taus
    if np.any(m):
        return np.argmin(m)
    return len(taus) - 1


def acl(x, c=5.0):
    """
    Compute autocorrelation length for x
    """
    N = x.shape[0]
    acf = autocorrelation(x)
    T = 2.0*np.cumsum(acf)-1.0
    return T[auto_window(T, c)]



import logging
from .utils import StreamHandler, LEVELS
# Configure base logger for the raynest package - inherited by all loggers with
# names prefixed by 'raynest'
logger = logging.getLogger('raynest')
logger.setLevel(LEVELS[-1])  # maximum verbosity recorded to base logger
logger.addHandler(logging.NullHandler())

console_handler = StreamHandler(verbose=0)  # default console verbosity is 0
logger.addHandler(console_handler)
# To change the console handler verbosity:
#   from raynest import console_handler
#   console_handler.set_verbosity(2)
__version__="1.0.2"

from .raynest import raynest

# Get the version number from git tag
from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = "dev"

__all__ = ['model',
           'NestedSampling',
           'parameter',
           'sampler',
           'raynest',
           'nest2pos',
           'proposal',
           'plot',
           'logger']



import numpy as np
import raynest.model

def sine_gaussian(x,t):
    e = (t-x['t0'])/x['tau']
    return x['A']*np.exp(-e**2)*np.cos(2*np.pi*x['f']*t + x['phi'])

class BurstModel(raynest.model.Model):
    """
    A Burst like search model
    """
    def __init__(self, time, data, sigma = 1.0):
        self.time   = time
        self.data   = data
        self.sigma  = sigma
        self.names  = ['A','f','tau','t0','phi']
        self.bounds = [[0.0,1.0],[10,100],[0.1,1.0],[time.min(),time.max()],[0.0,2*np.pi]]

    def log_likelihood(self,p):
        m = sine_gaussian(p, self.time)
        r = (self.data-m)/self.sigma
        return np.sum(-0.5*r**2)
    
    def log_prior(self,p):
        logP = super(BurstModel,self).log_prior(p)
        return logP
    
    def force(self, p):
        f = np.zeros(1, dtype = {'names':p.names, 'formats':['f8' for _ in p.names]})
        return f
    
    def analytical_gradient(self, p):
        return p.values

if __name__ == "__main__":
    np.random.seed(12)
    time  = np.linspace(0.0,1.0,1000)
    sigma = 0.1
    noise = np.random.normal(0,sigma,size = time.shape[0])
    truth = {'A':0.1,'f':50,'tau':0.05,'t0':0.5,'phi':1.0}
    signal = sine_gaussian(truth,time)
    data = noise+signal
    model=BurstModel(time, data, sigma=sigma)
    work=raynest.raynest(model, verbose=2, nnest=4, nensemble=8, nlive=1000, maxmcmc=2000, nslice=0, nhamiltonian=0, resume=1, periodic_checkpoint_interval=100, output='burst_test')
    work.run()
    import matplotlib.pyplot as plt
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    ax.plot(time,signal,lw=0.75,color='r',zorder=100)
    ax.plot(time,data,lw=0.5,color='k')
    pos = work.posterior_samples
    models = np.array([sine_gaussian(p,time) for p in pos])
    l,m,h  = np.percentile(models,[5,50,95],axis=0)
    ax.fill_between(time,l,h,facecolor='turquoise',alpha=0.5)
    ax.plot(time,m,lw=0.75,color='turquoise')
    ax.set_xlabel('time [s]')
    ax.set_ylabel('strain')
    plt.savefig('burst_test/burst_waveform.pdf',bbox_inches='tight')



import pstats, cProfile

s = pstats.Stats("prof_sampler.prof")
s.strip_dirs().sort_stats("time").print_stats()

s = pstats.Stats("prof_nested_sampling.prof")
s.strip_dirs().sort_stats("cumulative").print_stats()



import unittest
import numpy as np
import raynest.model

class GaussianModel(raynest.model.Model):
    """
    A simple gaussian model with parameters mean and sigma
    """
    names=['mean','sigma']
    bounds=[[-5,5],[0.05,1]]
    data = np.array([x for x in np.random.normal(0.5,0.5,size=10)])
    analyticZ = np.log(0.05)

    @classmethod
    def log_likelihood(cls,x):
        return -0.5*x['mean']**2/x['sigma']**2 - np.log(x['sigma']) - 0.5*np.log(2.0*np.pi)

    def log_prior(self,p):
        if not self.in_bounds(p): return -np.inf
        return -np.log(p['sigma']) - np.log(10) - np.log(0.95)

    def force(self,x):
        return np.zeros(1, dtype = {'names':x.names, 'formats':['f8' for _ in x.names]})

class GaussianTestCase(unittest.TestCase):
    """
    Test the gaussian model
    """
    def setUp(self):
        self.model = GaussianModel()
        self.work=raynest.raynest(self.model,verbose=0,nensemble=2,nlive=100,maxmcmc=100)

    def test_run(self):
        self.work.run()
        print('Analytic evidence: {0}'.format(self.model.analyticZ))



def test_all():
    unittest.main(verbosity=2)

if __name__=='__main__':
    unittest.main(verbosity=2)




import unittest
import numpy as np
import raynest.model

class GaussianModel(raynest.model.Model):
    """
    A simple 2 dimensional gaussian
    """
    def __init__(self):
        pass
    names=['x','y']
    bounds=[[-10,10],[-10,10]]
    analytic_log_Z=0.0 - np.log(bounds[0][1]-bounds[0][0]) - np.log(bounds[1][1]-bounds[1][0])

    def log_likelihood(self,p):
        return -0.5*(p['x']**2 + p['y']**2) - np.log(2.0*np.pi)

    def log_prior(self,p):
        return super(GaussianModel,self).log_prior(p)
    
    def force(self, p):
        f = np.zeros(1, dtype = {'names':p.names, 'formats':['f8' for _ in p.names]})
        return f

class GaussianTestCase(unittest.TestCase):
    """
    Test the gaussian model
    """
    def setUp(self):
        gaussmodel = GaussianModel()
        self.work=raynest.raynest(gaussmodel,verbose=2,nensemble=2,nlive=1000,maxmcmc=500,nslice=0,nhamiltonian=0)
        print('Sampling 2D gaussian with analytic evidence {0}'.format(gaussmodel.analytic_log_Z))

    def test_run(self):
        self.work.run()
        # 2 sigma tolerance
        logZ = self.work.logZ
        H    = self.work.information
        tolerance = 2.0*self.work.logZ_error
        self.assertTrue(np.abs(logZ - GaussianModel.analytic_log_Z)<tolerance, 'Incorrect evidence for normalised distribution: {0} +/- {2} instead of {1}'.format(logZ ,GaussianModel.analytic_log_Z, tolerance))

def test_all():
    unittest.main(verbosity=2)

if __name__=='__main__':
    unittest.main(verbosity=2)
 



import unittest
import numpy as np
from scipy import integrate,stats
import raynest
import raynest.model
import matplotlib as mpl
mpl.use('Agg')
from matplotlib import pyplot as plt

class GaussianModel(raynest.model.Model):
    """
    A simple gaussian model with parameters mean and sigma
    """
    def __init__(self):
        self.distr = stats.norm(loc=0,scale=1.0)
    names=['x']
    bounds=[[-10,10]]
    analytic_log_Z=0.0 - np.log(bounds[0][1] - bounds[0][0])

    def log_likelihood(self,p):
        return self.distr.logpdf(p['x'])
        #return -0.5*(p['x']**2) - 0.5*np.log(2.0*np.pi)
        
    def force(self, p):
        f = np.zeros(1, dtype = {'names':p.names, 'formats':['f8' for _ in p.names]})
        return f

class GaussianTestCase(unittest.TestCase):
    """
    Test the gaussian model
    """
    def setUp(self):
        self.model=GaussianModel()
        self.work=raynest.raynest(self.model,verbose=1,nlive=1000,nensemble=2,maxmcmc=200)
        self.work.run()

    def test_evidence(self):
        # 2 sigma tolerance
        logZ = self.work.logZ
        H    = self.work.information
        tolerance = 2.0*self.work.logZ_error

        print('2-sigma statistic error in logZ: {0:0.3f}'.format(tolerance))
        print('Analytic logZ {0}'.format(self.model.analytic_log_Z))
        print('Estimated logZ {0}'.format(logZ))
        pos=self.work.posterior_samples['x']
        #t,pval=stats.kstest(pos,self.model.distr.cdf)
        stat,pval = stats.normaltest(pos.T)
        print('Normal test p-value {0}'.format(str(pval)))
        plt.figure()
        plt.hist(pos.ravel(),density=True)
        x=np.linspace(self.model.bounds[0][0],self.model.bounds[0][1],100)
        plt.plot(x,self.model.distr.pdf(x))
        plt.title('NormalTest pval = {0}'.format(pval))
        plt.savefig('posterior.png')
        plt.figure()
        plt.plot(pos.ravel(),',')
        plt.title('chain')
        plt.savefig('chain.png')
        self.assertTrue(np.abs(logZ - GaussianModel.analytic_log_Z)<tolerance, 'Incorrect evidence for normalised distribution: {0:.3f} instead of {1:.3f}'.format(logZ,GaussianModel.analytic_log_Z ))
        self.assertTrue(pval>0.01,'Normaltest test failed: KS stat = {0}'.format(pval))



def test_all():
    unittest.main(verbosity=2)

if __name__=='__main__':
    unittest.main(verbosity=2)




import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import sys, os

np.seterr(all='raise')

def log_likelihood(x):
    return np.sum([-0.5*x[n]**2-0.5*np.log(2.0*np.pi) for n in range(x.shape[0])])

mode = sys.argv[1]
if mode == 'delete':
    allfiles = os.listdir('.')
    toremove = [a for a in allfiles if 'trajectory_' in a and '.py' not in a]
    for f in toremove: os.remove(f)
    exit()
traj = np.genfromtxt('trajectory_'+sys.argv[1]+'.txt', names= True)
npts = 256
x = np.linspace(-10,10,npts)
y = np.linspace(-10,10,npts)
X, Y = np.meshgrid(x, y)
Z   = np.zeros((npts,npts))
for i in range(npts):
    for j in range(npts):
        Z[i,j] = log_likelihood(np.array([x[i],y[j]]))

C = plt.contour(X, Y, Z, levels = [traj['logLmin'][0]], linewidths=1.0,colors='k')
plt.contourf(X, Y, Z, 6, cmap = cm.Greys_r)
S = plt.scatter(traj['0'], traj['1'], c=traj['logL'], s = 8)
plt.plot(traj['0'],traj['1'], color = 'k', lw = 0.5)
for k in range(traj.shape[0]):
    plt.text(traj['0'][k],traj['1'][k], str(k), color="black", fontsize=8)
plt.colorbar(S)
plt.show()



import sys
import unittest
import numpy as np
from scipy import integrate,stats
import raynest
import raynest.model
import matplotlib as mpl
mpl.use('Agg')
from matplotlib import pyplot as plt

class HalfGaussianModel(raynest.model.Model):
    """
    A simple gaussian model with parameters mean and sigma
    """
    def __init__(self,xmax=10.0):
        self.distr = stats.norm(loc=0,scale=1.0)
        self.bounds=[[0,xmax]]
        self.names=['x']
        self.analytic_log_Z = np.log(self.distr.cdf(self.bounds[0][1]) - self.distr.cdf(self.bounds[0][0])) \
                - np.log(self.bounds[0][1]-self.bounds[0][0])

    def log_likelihood(self,p):
        return self.distr.logpdf(p['x'])

    def force(self,x):
        return np.zeros(1, dtype = {'names':x.names, 'formats':['f8' for _ in x.names]})

class HalfGaussianTestCase(unittest.TestCase):
    """
    Test the gaussian model
    """
    def setUp(self,xmax=20):
        self.model=HalfGaussianModel(xmax=xmax)
        self.work=raynest.raynest(self.model,verbose=1,nlive=500,nensemble=1)
        self.work.run()

    def test_evidence(self):
        # 2 sigma tolerance
        logZ = self.work.logZ
        H    = self.work.information
        tolerance = 2.0*self.work.logZ_error
        print('2-sigma statistic error in logZ: {0:0.3f}'.format(tolerance))
        print('Analytic logZ {0}'.format(self.model.analytic_log_Z))
        print('Estimated logZ {0}'.format(logZ))
        pos=self.work.posterior_samples['x']

        plt.figure()
        plt.hist(pos.ravel(),density=True)
        x=np.linspace(self.model.bounds[0][0],self.model.bounds[0][1],100)
        plt.plot(x,2*self.model.distr.pdf(x))
        plt.savefig('posterior.png')
        plt.figure()
        plt.plot(pos.ravel(),',')
        plt.title('chain')
        plt.savefig('chain.png')
        self.assertTrue(np.abs(logZ - self.model.analytic_log_Z)<tolerance, 'Incorrect evidence for normalised distribution: {0:.3f} instead of {1:.3f}'.format(logZ,self.model.analytic_log_Z ))


def test_all():
    unittest.main(verbosity=2)

if __name__=='__main__':
    unittest.main()




#!/bin/usr/env python

import numpy as np
import raynest.model

class RosenbrockModel(raynest.model.Model):
    """
    The n-dimensional Rosenbrock function

    See: https://arxiv.org/pdf/1903.09556.pdf
    """
    def __init__(self, ndims=2):
        self.ndims = ndims
        self.names = ['x' + str(i) for i in range(ndims)]
        self.bounds = [[-5, 5] for i in range(ndims)]


    def log_likelihood(self, x):
        x = np.array(x.values).reshape(-1, self.ndims)
        return  -(np.sum(100. * (x[:,1:] - x[:,:-1] ** 2.) ** 2. + (1. - x[:,:-1]) ** 2., axis=1))

    def force(self, x):
        f = np.zeros(1, dtype = {'names':x.names, 'formats':['f8' for _ in x.names]})
        return f

if __name__=='__main__':
    model=RosenbrockModel(ndims = 2)
    work=raynest.raynest(model,
                         verbose=2,
                         nnest=2,
                         nensemble=2,
                         nlive=100,
                         maxmcmc=5000,
                         nslice=0,
                         nhamiltonian=0,
                         seed = 1,
                         resume=0,
                         output='rosenbrock',
                         periodic_checkpoint_interval=600)
    work.run(corner = True)




import unittest
import numpy as np
import raynest.model
from scipy import stats

class GaussianModel(raynest.model.Model):
    """
    An n-dimensional gaussian
    """
    def __init__(self,dim=50):
        self.distr = stats.norm(loc=0,scale=1.0)
        self.dim=dim
        self.names=['{0}'.format(i) for i in range(self.dim)]
        self.bounds=[[-10,10] for _ in range(self.dim)]
#        self.bounds[0] = [-50, 50]
        self.analytic_log_Z=0.0 - sum([np.log(self.bounds[i][1]-self.bounds[i][0]) for i in range(self.dim)])

    def log_likelihood(self,p):
        return np.sum([-0.5*p[n]**2-0.5*np.log(2.0*np.pi) for n in p.names])##np.sum([self.distr.logpdf(p[n]
    
    def log_prior(self,p):
        logP = super(GaussianModel,self).log_prior(p)
#        for n in p.names: logP += -0.5*(p[n])**2
        return logP
    
    def force(self, p):
        f = np.zeros(1, dtype = {'names':p.names, 'formats':['f8' for _ in p.names]})
        return f
    
    def analytical_gradient(self, p):
        return p.values

class GaussianTestCase(unittest.TestCase):
    """
    Test the gaussian model
    """
    def setUp(self):
        self.model=GaussianModel(dim = 50)
        self.work=raynest.raynest(self.model, verbose=2, nensemble=0, nlive=1000, maxmcmc=5000, nslice=6, nhamiltonian=0, resume=0)

    def test_run(self):
        self.work.run()
        # 2 sigma tolerance
        tolerance = 2.0*np.sqrt(self.work.NS.info/self.work.NS.nlive)
        self.assertTrue(np.abs(self.work.NS.logZ - self.model.analytic_log_Z)<tolerance, 'Incorrect evidence for normalised distribution: {0} instead of {1}'.format(self.work.NS.logZ ,self.model.analytic_log_Z))
        print("analytic logZ = {0}".format(self.model.analytic_log_Z))
def test_all():
    unittest.main(verbosity=2)

if __name__=='__main__':
#    unittest.main(verbosity=2)
    model=GaussianModel(dim = 50)
    work=raynest.raynest(model, verbose=2,
                       nnest=1, nensemble=3, nlive=1000, maxmcmc=5000, nslice=0, nhamiltonian=0, seed = 1,
                       resume=0, periodic_checkpoint_interval=600)
    work.run()
    print("analytic logZ = {0}".format(model.analytic_log_Z))
    print("estimated logZ = {0} \pm {1}".format(work.logZ,work.logZ_error))



import unittest
import numpy as np
import raynest.model

class EggboxModel(raynest.model.Model):
    """
    Eggbox problem from https://arxiv.org/pdf/0809.3437v1.pdf
    """
    names=['1','2','3','4','5']
    bounds=[[0,10.0*np.pi],[0,10.0*np.pi],[0,10.0*np.pi],[0,10.0*np.pi],[0,10.0*np.pi]]
    data = None
    
    def log_likelihood(self,x):
        return log_eggbox(x)
    
    def log_prior(self,x):
        return super(EggboxModel,self).log_prior(x)

def log_eggbox(p):
    tmp = 1.0
    for n in p.names:
        tmp *= np.cos(p[n]/2.)
    return (tmp+2.0)**5.0

if __name__=='__main__':
    M = EggboxModel()
    work=raynest.raynest(M,verbose=2,nnest=4,nensemble=4,nlive=1000,maxmcmc=1000,nslice=4)
    work.run()




from __future__ import division
from functools import reduce
import numpy as np
from math import log,sqrt,fabs,exp
from abc import ABCMeta,abstractmethod
from scipy.stats import multivariate_normal
from scipy.special import logsumexp
from .nest2pos import acl
from .parameter import LivePoint
from .NestedSampling import LivePoints
import ray
    
class Proposal(object):
    """
    Base abstract class for jump proposals
    """
    __metaclass__ = ABCMeta
    log_J = 0.0 # Jacobian of this jump proposal
    rng   = None # numpy Generator object
    
    def __init__(self, rng, *args, **kwargs):
        self.rng = rng
    
    @abstractmethod
    def get_sample(self,old):
        """
        Returns a new proposed sample given the old one.
        Must be implemented by user

        Parameters
        ----------
        old : :obj:`raynest.parameter.LivePoint`

        Returns
        ----------
        out: :obj:`raynest.parameter.LivePoint`
        """
        pass

class EnsembleProposal(Proposal):
    """
    Base class for ensemble proposals
    """
    ensemble=None
    
    def set_ensemble(self,ensemble):
        """
        Set the ensemble of points to use
        """
        self.ensemble = ensemble
        
class ProposalCycle(EnsembleProposal):
    """
    A proposal that cycles through a list of
    jumps.

    Initialisation arguments:

    proposals : A list of jump proposals
    weights   : Weights for each type of jump

    Optional arguments:
    cyclelength : length of the proposal cycle. Default: 100

    """
    idx=0 # index in the cycle
    N=0   # number of proposals in the cycle
    def __init__(self,proposals,weights,rng,cyclelength=100,*args,**kwargs):
        super(ProposalCycle,self).__init__(rng)
        assert(len(weights)==len(proposals))
        self.rng         = rng
        self.cyclelength = cyclelength
        self.weights     = weights
        self.proposals   = proposals
        self.set_cycle()

    def set_cycle(self):
        # The cycle is a list of indices for self.proposals
        self.cycle = self.rng.choice(self.proposals, size=self.cyclelength,
                                     p=self.weights, replace=True)
        self.N=len(self.cycle)

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, weights):
        self._weights = self.normalise_weights(weights)

    def normalise_weights(self, weights):
        norm = sum(weights)
        for i, _ in enumerate(weights):
            weights[i]=weights[i] / norm
        return weights

    def get_sample(self,old,**kwargs):
        # Call the current proposal and increment the index
        self.idx = (self.idx + 1) % self.N
        p = self.cycle[self.idx]
        new = p.get_sample(old,**kwargs)
        self.log_J = p.log_J
        return new

    def set_ensemble(self,ensemble):
        """
        Updates the ensemble statistics
        by calling it on each :obj:`EnsembleProposal`
        """
        for p in self.proposals:
            if isinstance(p,EnsembleProposal):
                p.set_ensemble(ensemble)
        del ensemble

    def add_proposal(self, proposal, weight):
        self.proposals = self.proposals + [proposal]
        self.weights = self.weights + [weight]
        self.set_cycle()

class EnsembleSlice(EnsembleProposal):
    """
    The Ensemble Slice proposal from Karamanis & Beutler
    https://arxiv.org/pdf/2002.06212v1.pdf
    """
    log_J      = 0.0 # Symmetric proposal
    mean       = None
    covariance = None
            
    def set_ensemble(self,ensemble):
        """
        Over-ride default set_ensemble so that the
        mean and covariance matrix are recomputed when it is updated
        """
        super(EnsembleSlice,self).set_ensemble(ensemble)
        self.mean, self.covariance = ensemble.get_mean_covariance()
        
class EnsembleSliceDifferential(EnsembleSlice):
    """
    The Ensemble Slice Differential move from Karamanis & Beutler
    https://arxiv.org/pdf/2002.06212v1.pdf
    """

    def get_direction(self, mu = 1.0):
        """
        Draws two random points and returns their direction
        """
        subset = self.ensemble.sample(2)
        direction = reduce(LivePoint.__sub__,subset)
        return direction * mu

class EnsembleSliceCorrelatedGaussian(EnsembleSlice):
    """
    The Ensemble Slice Correlated Gaussian move from Karamanis & Beutler
    https://arxiv.org/pdf/2002.06212v1.pdf
    """
    def get_direction(self, mu = 1.0):
        """
        Draws a random gaussian direction
        """
        direction = mu * self.rng.multivariate_normal(self.mean, self.covariance)
        return direction

class EnsembleSliceGaussian(EnsembleSlice):
    """
    The Ensemble Slice Gaussian move from Karamanis & Beutler
    https://arxiv.org/pdf/2002.06212v1.pdf
    """

    def get_direction(self, mu = 1.0):
        """
        Draw a random gaussian direction
        """
        direction  = self.rng.normal(0.0,1.0,size=len(self.mean))
        direction /= np.linalg.norm(direction)
        return direction * mu

class EnsembleSliceProposalCycle(ProposalCycle):
    def __init__(self, rng, model=None):
        """
        A proposal cycle that uses the slice sampler :obj:`EnsembleSlice`
        proposal.
        """
        weights = [1,1,1]
        proposals = [EnsembleSliceDifferential(rng),EnsembleSliceGaussian(rng),EnsembleSliceCorrelatedGaussian(rng)]
        super(EnsembleSliceProposalCycle,self).__init__(proposals, weights, rng)

    def get_direction(self, mu = 1.0, **kwargs):
        """
        Get a direction for the slice jump
        """
        self.idx = (self.idx + 1) % self.N
        p = self.cycle[self.idx]
        new = p.get_direction(mu = mu, **kwargs)
        self.log_J = p.log_J
        return new

class EnsembleWalk(EnsembleProposal):
    """
    The Ensemble "walk" move from Goodman & Weare
    http://dx.doi.org/10.2140/camcos.2010.5.65

    Draws a step by evolving along the
    direction of the center of mass of
    3 points in the ensemble.
    """
    log_J = 0.0 # Symmetric proposal
    Npoints = 3
    def get_sample(self,old):
        """
        Parameters
        ----------
        old : :obj:`raynest.parameter.LivePoint`

        Returns
        ----------
        out: :obj:`raynest.parameter.LivePoint`
        """
        subset = self.ensemble.sample(self.Npoints)
        center_of_mass = reduce(type(old).__add__,subset)/float(self.Npoints)
        out = old
        for x in subset:
            out += (x - center_of_mass)*self.rng.standard_normal()
        return out

class EnsembleModifiedWalk(EnsembleProposal):
    """
    The Ensemble "walk" move from Goodman & Weare
    http://dx.doi.org/10.2140/camcos.2010.5.65
    modified as in https://arxiv.org/pdf/2411.00276
    """
    log_J = 0.0 # Symmetric proposal
    Npoints = 3
    tj = -1.0
    tk = +1.0
    
    def lagrange_weights(self, x, x0, x1, x2):
        return (x-x1)/(x0-x1) * (x-x2)/(x0-x2)
        
    def get_sample(self,old):
        """
        Parameters
        ----------
        old : :obj:`raynest.parameter.LivePoint`

        Returns
        ----------
        out: :obj:`raynest.parameter.LivePoint`
        """
        subset_1 = self.ensemble.sample(self.Npoints)
        center_of_mass_1 = reduce(type(old).__add__,subset_1)/float(self.Npoints)
        subset_2 = self.ensemble.sample(self.Npoints)
        center_of_mass_2 = reduce(type(old).__add__,subset_2)/float(self.Npoints)
        
        ti   = self.rng.standard_normal()
        tNew = self.rng.standard_normal()
        
        wi = self.lagrange_weights(tNew,ti,self.tj,self.tk);
        wj = self.lagrange_weights(tNew,self.tj,self.tk,ti);
        wk = self.lagrange_weights(tNew,self.tk,ti,self.tj);
        
        out = old.copy()
        out.values = wi*old.values + wj*center_of_mass_1.values + wk*center_of_mass_2.values
        # Jacobian
        self.log_J = out.dimension * np.log(np.abs(wi))
        return out

class EnsembleStretch(EnsembleProposal):
    """
    The Ensemble "stretch" move from Goodman & Weare
    http://dx.doi.org/10.2140/camcos.2010.5.65
    """
    def get_sample(self,old):
        """
        Parameters
        ----------
        old : :obj:`raynest.parameter.LivePoint`

        Returns
        ----------
        out: :obj:`raynest.parameter.LivePoint`
        """
        scale = 2.0 # Will stretch factor in (1/scale,scale)
        # Pick a random point to move toward
        a = self.ensemble.sample(1)[0]
        # Pick the scale factor
        x = self.rng.uniform(-1,1)*np.log(scale)
        Z = exp(x)
        out = a + (old - a)*Z
        # Jacobian
        self.log_J = out.dimension * x
        return out

class EnsembleModifiedStretch(EnsembleProposal):
    """
    The Ensemble "stretch" move from Goodman & Weare
    http://dx.doi.org/10.2140/camcos.2010.5.65
    modified as in https://arxiv.org/pdf/2411.00276
    """
    def get_sample(self,old):
        """
        Parameters
        ----------
        old : :obj:`raynest.parameter.LivePoint`

        Returns
        ----------
        out: :obj:`raynest.parameter.LivePoint`
        """
        scale  = 2.0 # Will stretch factor in (1/scale,scale)
        # Pick a random point to move toward
        a, b   = self.ensemble.sample(2)
        c      = self.rng.uniform(-1,1)
        r_star = c*a.values+(1-c)*b.values
        # Pick the scale factor
        x = self.rng.uniform(-1,1)*np.log(scale)
        Z = exp(x)
        out = old.copy()
        out.values = Z*old.values + (1-Z)*r_star
        # Jacobian
        self.log_J = out.dimension * x
        return out
        
class DifferentialEvolution(EnsembleProposal):
    """
    Differential evolution move:
    Draws a step by taking the difference vector between two points in the
    ensemble and adding it to the current point.
    See e.g. Exercise 30.12, p.398 in MacKay's book
    http://www.inference.phy.cam.ac.uk/mackay/itila/

    We add a small perturbation around the exact step
    """
    log_J = 0.0 # Symmetric jump
    def get_sample(self,old):
        """
        Parameters
        ----------
        old : :obj:`raynest.parameter.LivePoint`

        Returns
        ----------
        out: :obj:`raynest.parameter.LivePoint`
        """
        a, b = self.ensemble.sample(2)
        sigma = 1e-4 # scatter around difference vector by this factor
        out = old + (b-a)*self.rng.normal(1.0,sigma)
        return out

class EnsembleEigenVector(EnsembleProposal):
    """
    A jump along a randomly-chosen eigenvector
    of the covariance matrix of the ensemble
    """
    log_J         = 0.0
    eigen_values  = None
    eigen_vectors = None
    covariance    = None
    ensemble      = None
    def set_ensemble(self, ensemble):
        """
        Over-ride default set_ensemble so that the
        eigenvectors are recomputed when it is updated
        """
        self.update_eigenvectors(ensemble)

    def update_eigenvectors(self, ensemble):
        """
        Recompute the eigenvectors and eigevalues
        of the covariance matrix of the ensemble
        """
        self.eigen_values, self.eigen_vectors = ensemble.get_eigen_quantities()

    def get_sample(self,old):
        """
        Propose a jump along a random eigenvector
        Parameters
        ----------
        old : :obj:`raynest.parameter.LivePoint`

        Returns
        ----------
        out: :obj:`raynest.parameter.LivePoint`
        """
        out = old.copy()
        # pick a random eigenvector
        i = self.rng.integers(low=0, high=old.dimension)
        jumpsize = sqrt(fabs(self.eigen_values[i]))*self.rng.standard_normal()
        out.values += jumpsize*self.eigen_vectors[:,i]
        return out

class EnsembleQuadratic(EnsembleProposal):

    tj = -1.0
    tk = +1.0
    
    def lagrange_weights(self, x, x0, x1, x2):
        return (x-x1)/(x0-x1) * (x-x2)/(x0-x2)

    def get_sample(self,old):
        """
        Propose a jump using the quadratic move in Militzer (2023a)
        ----------
        old : :obj:`raynest.parameter.LivePoint`

        Returns
        ----------
        out: :obj:`raynest.parameter.LivePoint`
        """
        a, b = self.ensemble.sample(2)

        ti   = self.rng.standard_normal()
        tNew = self.rng.standard_normal()
        
        wi = self.lagrange_weights(tNew,ti,self.tj,self.tk);
        wj = self.lagrange_weights(tNew,self.tj,self.tk,ti);
        wk = self.lagrange_weights(tNew,self.tk,ti,self.tj);
        out = old.copy()
        out.values = wi*old.values + wj*a.values + wk*b.values
        # Jacobian
        self.log_J = out.dimension * np.log(np.abs(wi))
        return out

class DefaultProposalCycle(ProposalCycle):
    """
    A default proposal cycle that uses the
    :obj:`raynest.proposal.EnsembleWalk`, :obj:`raynest.proposal.EnsembleStretch`,
    :obj:`raynest.proposal.DifferentialEvolution`, :obj:`raynest.proposal.EnsembleEigenVector`
    ensemble proposals.
    """
    def __init__(self, rng, *args, **kwargs):

        proposals = [EnsembleWalk(rng),
                     EnsembleModifiedWalk(rng),
                     EnsembleStretch(rng),
                     EnsembleModifiedStretch(rng),
                     EnsembleQuadratic(rng),
                     DifferentialEvolution(rng),
                     EnsembleEigenVector(rng)]
        weights = [3,
                   1,
                   3,
                   1,
                   5,
                   3,
                   10]

        super(DefaultProposalCycle,self).__init__(proposals, weights, rng)

class HamiltonianProposalCycle(ProposalCycle):
    def __init__(self, rng, model=None, density=None):
        """
        A proposal cycle that uses the hamiltonian :obj:`ConstrainedLeapFrog`
        proposal.
        Requires a :obj:`raynest.Model` to be passed for access to the user-defined
        :obj:`raynest.Model.force` (the gradient of :obj:`raynest.Model.potential`) and
        :obj:`raynest.Model.log_likelihood` to define the reflective
        """
        weights = [1]
        proposals = [ConstrainedLeapFrog(rng, model=model, density=density)]
        super(HamiltonianProposalCycle,self).__init__(proposals, weights, rng)

class HamiltonianProposal(EnsembleProposal):
    """
    Base class for hamiltonian proposals
    """
    covariance           = None
    mass_matrix          = None
    inverse_mass_matrix  = None
    momenta_distribution = None

    def __init__(self, rng, model=None, density=None):
        """
        Initialises the class with the kinetic
        energy and the :obj:`raynest.Model.potential`.
        """
        super(HamiltonianProposal, self).__init__(rng)
        self.rng                    = rng
        self.T                      = self.kinetic_energy
        self.model                  = model
        self.V                      = self.model.potential
        self.prior_bounds           = self.model.bounds
        self.dimension              = len(self.prior_bounds)
        self.dt                     = 0.3
        self.leaps                  = 100
        self.maxleaps               = 1000
        self.DEBUG                  = 0
        self.trajectories           = []
        self.covariance             = np.identity(self.dimension)
        
        self.set_mass_parameters()
        self.set_momenta_distribution()
    
    def set_mass_parameters(self):
        self.mass_matrix         = self.covariance
        self.inverse_mass_matrix = np.linalg.inv(self.mass_matrix)
        self.inverse_mass        = np.atleast_1d(np.squeeze(np.diag(self.inverse_mass_matrix)))
        _, self.logdeterminant   = np.linalg.slogdet(self.mass_matrix)
        
    def set_ensemble(self, density_covariance):
        """
        FIXME:
        """
        self.density = density_covariance[0]
#        self.covariance = density_covariance[1]
#        self.set_mass_parameters()

    def unit_normal(self, q):
        v = self.density._log_gradient(q.values)
        return v/np.linalg.norm(v)

    def force(self, q):
        """
        return the gradient of the potential function as numpy ndarray
        Parameters
        ----------
        q : :obj:`raynest.parameter.LivePoint`
            position

        Returns
        ----------
        dV: :obj:`numpy.ndarray` gradient evaluated at q
        """
        return -self.density._log_gradient(q.values)

    def set_momenta_distribution(self):
        """
        update the momenta distribution using the
        mass matrix (precision matrix of the ensemble).
        """
        self.momenta_distribution = multivariate_normal(cov=self.mass_matrix)
        
    def kinetic_energy(self,p):
        """
        kinetic energy part for the Hamiltonian.
        Parameters
        ----------
        p : :obj:`numpy.ndarray`
            momentum

        Returns
        ----------
        T: :float: kinetic energy
        """
        return 0.5 * np.dot(p,np.dot(self.inverse_mass_matrix,p))-self.logdeterminant

    def hamiltonian(self, p, q):
        """
        Hamiltonian.
        Parameters
        ----------
        p : :obj:`numpy.ndarray`
            momentum
        q : :obj:`raynest.parameter.LivePoint`
            position
        Returns
        ----------
        H: :float: hamiltonian
        """
        return self.T(p) + self.V(q)

class LeapFrog(HamiltonianProposal):
    """
    Leap frog integrator proposal for an unconstrained
    Hamiltonian Monte Carlo step
    """
    def __init__(self, *args, model=None, density=None):
        """
        Parameters
        ----------
        model : :obj:`raynest.Model`
        """
        super(LeapFrog, self).__init__(model=model, density=density)
        self.prior_bounds   = model.bounds

    def get_sample(self, q0, *args):
        """
        Propose a new sample, starting at q0

        Parameters
        ----------
        q0 : :obj:`raynest.parameter.LivePoint`
            position

        Returns
        ----------
        q: :obj:`raynest.parameter.LivePoint`
            position
        """
        # compute local equilibrium distribution
#        self.set_lte(q0)
        # generate a canonical momentum
        p0 = np.atleast_1d(self.momenta_distribution.rvs())
        initial_energy = self.hamiltonian(p0,q0)
        # evolve along the trajectory
        q, p, r = self.evolve_trajectory(p0, q0, *args)
        # minus sign from the definition of the potential
        final_energy   = self.hamiltonian(p, q)
        if r == 1:
            self.log_J = -np.inf
        else:
            self.log_J = min(0.0, initial_energy-final_energy)
        return q

    def evolve_trajectory(self, p0, q0, *args):
        """
        Hamiltonian leap frog trajectory subject to the
        hard boundary defined by the parameters prior bounds.
        https://arxiv.org/pdf/1206.1901.pdf

        Parameters
        ----------
        p0 : :obj:`numpy.ndarray`
            momentum
        q0 : :obj:`raynest.parameter.LivePoint`
            position

        Returns
        ----------
        p: :obj:`numpy.ndarray` updated momentum vector
        q: :obj:`raynest.parameter.LivePoint`
            position
        """
        # Updating the momentum a half-step
        p = p0 - 0.5 * self.dt * self.gradient(q0)
        q = q0.copy()
        
        for i in range(self.leaps):

            # do a step
            for j,k in enumerate(q.names):
                u,l = self.prior_bounds[j][1], self.prior_bounds[j][0]
                q[k] += self.dt * p[j] * self.inverse_mass[j]
                # check and reflect against the bounds
                # of the allowed parameter range
                while q[k] <= l or q[k] >= u:
                    if q[k] > u:
                        q[k] = u - (q[k] - u)
                        p[j] *= -1
                    if q[k] < l:
                        q[k] = l + (l - q[k])
                        p[j] *= -1

            dV = self.gradient(q)
            # take a full momentum step
            p += - self.dt * dV
        # Do a final update of the momentum for a half step
        p += - 0.5 * self.dt * dV

        return q, -p, 0

class ConstrainedLeapFrog(HamiltonianProposal):
    """
    Leap frog integrator proposal for a costrained
    (logLmin defines a reflective boundary)
    Hamiltonian Monte Carlo step.
    """
    def __init__(self, *args, model=None, density=None):
        """
        Parameters
        ----------
        model : :obj:`raynest.Model`
        """
        super(ConstrainedLeapFrog, self).__init__(*args, model=model, density=density)
        self.log_likelihood = model.log_likelihood
        self.density        = density
        
    def get_sample(self, q0, logLmin=-np.inf):
        """
        Generate new sample with constrained HMC, starting at q0.

        Parameters
        ----------
        q0 : :obj:`raynest.parameter.LivePoint`
            position

        logLmin: hard likelihood boundary

        Returns
        ----------
        q: :obj:`raynest.parameter.LivePoint`
            position
        """
        p0 = np.atleast_1d(self.momenta_distribution.rvs())
        initial_energy = self.hamiltonian(p0,q0)
        # evolve along the trajectory
        q, p, r = self.evolve_trajectory(p0, q0, logLmin)
        # minus sign from the definition of the potential
        final_energy   = self.hamiltonian(p, q)
        if r == 1:
            self.log_J = -np.inf
        else:
            self.log_J = min(0.0, initial_energy-final_energy)
        return q

    def counter(self):
        n = 0
        while True:
            yield n
            n += 1

    def evolve_trajectory_one_step_position(self, p, q):
        """
        One leap frog step in position

        Parameters
        ----------
        p0 : :obj:`numpy.ndarray`
            momentum
        q0 : :obj:`raynest.parameter.LivePoint`
            position
        Returns
        ----------
        p: :obj:`numpy.ndarray` updated momentum vector
        q: :obj:`raynest.parameter.LivePoint` position
        """
        for j,k in enumerate(q.names):
            u, l  = self.prior_bounds[j][1], self.prior_bounds[j][0]
            q[k]  += self.dt * p[j] * self.inverse_mass[j]
            # check and reflect against the bounds
            # of the allowed parameter range
            while q[k] < l or q[k] > u:
                if q[k] > u:
                    q[k] = u - (q[k] - u)
                    p[j] *= -1
                if q[k] < l:
                    q[k] = l + (l - q[k])
                    p[j] *= -1
        return p, q

    def evolve_trajectory_one_step_momentum(self, p, q, logLmin, half = False):
        """
        One leap frog step in momentum

        Parameters
        ----------
        p0 : :obj:`numpy.ndarray`
            momentum
        q0 : :obj:`raynest.parameter.LivePoint`
            position
        logLmin: :obj:`numpy.float64`
            loglikelihood constraint
        Returns
        ----------
        p: :obj:`numpy.ndarray` updated momentum vector
        q: :obj:`raynest.parameter.LivePoint` position
        """
        reflected = 0
        dV        = self.force(q)
        if half is True:
            p += - 0.5 * self.dt * dV
            return p, q, reflected
        else:
            c = self.check_constraint(q, logLmin)
            if c > 0:
                p += - self.dt * dV
            else:
                normal = self.unit_normal(q)
                p += - 2.0*np.dot(p,normal)*normal
                reflected = 1
        return p, q, reflected

    def check_constraint(self, q, logLmin):
        """
        Check the likelihood

        Parameters
        ----------
        q0 : :obj:`raynest.parameter.LivePoint`
        position
        logLmin: :obj:`numpy.float64`
        loglikelihood constraint
        Returns
        ----------
        c: :obj:`numpy.float64` value of the constraint
        """
        q.logP  = -self.V(q)
        q.logL  = self.log_likelihood(q)
        return q.logL - logLmin

    def evolve_trajectory(self, p0, q0, logLmin):
        """
        Evolve point according to Hamiltonian method in
        https://arxiv.org/pdf/1005.0157.pdf

        Parameters
        ----------
        p0 : :obj:`numpy.ndarray`
            momentum
        q0 : :obj:`raynest.parameter.LivePoint`
            position

        Returns
        ----------
        p: :obj:`numpy.ndarray` updated momentum vector
        q: :obj:`raynest.parameter.LivePoint` position
        """
        trajectory = [(q0,p0,0)]
        # evolve forward in time
        i = 0
        p, q, reflected = self.evolve_trajectory_one_step_momentum(p0.copy(), q0.copy(), logLmin, half = True)
        while (i < self.leaps):
            p, q            = self.evolve_trajectory_one_step_position(p, q)
            p, q, reflected = self.evolve_trajectory_one_step_momentum(p, q, logLmin, half = False)
            trajectory.append((q.copy(),p.copy(),reflected))
            i += 1

        p, q, reflected     = self.evolve_trajectory_one_step_momentum(p, q, logLmin, half = True)
        if self.DEBUG: self.save_trajectory(trajectory, logLmin)

#        # evolve backward in time
#        i = 0
#        p, q, reflected = self.evolve_trajectory_one_step_momentum(-p0.copy(), q0.copy(), logLmin, half = True)
#        while (i < self.leaps//2):
#            p, q            = self.evolve_trajectory_one_step_position(p, q)
#            p, q, reflected = self.evolve_trajectory_one_step_momentum(p, q, logLmin, half = False)
#            trajectory.append((q.copy(),p.copy(),reflected))
#            i += 1
#
#        p, q, reflected     = self.evolve_trajectory_one_step_momentum(p, q, logLmin, half = True)
#
#        if self.DEBUG: self.save_trajectory(trajectory, logLmin)
#        q, p, reflected = self.sample_trajectory(trajectory)
#
#        self.trajectories.append(trajectory)
        return q, -p, reflected

    def sample_trajectory(self, trajectory):
        """

        """
        logw = np.array([-self.hamiltonian(p,q) for q,p,_ in trajectory[1:-1]])
        norm = logsumexp(logw)
        idx  = self.rng.choice(range(1,len(trajectory)-1), p = np.exp(logw  - norm))
        return trajectory[idx]

    def save_trajectory(self, trajectory, logLmin, filename = None):
        """
        save trajectory for diagnostic purposes
        """
        if filename is None:
            filename = 'trajectory_'+str(next(self.c))+'.txt'
        f = open(filename,'w')
        names = trajectory[0][0].names

        for n in names:
            f.write(n+'\t'+'p_'+n+'\t')
        f.write('logPrior\tlogL\tlogLmin\n')

        for j,step in enumerate(trajectory):
            q = step[0]
            p = step[1]
            for j,n in enumerate(names):
                f.write(repr(q[n])+'\t'+repr(p[j])+'\t')
            f.write(repr(q.logP)+'\t'+repr(q.logL)+'\t'+repr(logLmin)+'\n')
        f.close()
        if self.c == 3: exit()

class NoUTurn(ConstrainedLeapFrog):
    def build_tree(theta, r, grad, v, j, epsilon, f, joint0):
        """The main recursion."""
        if (j == 0):
            # Base case: Take a single leapfrog step in the direction v.
            thetaprime, rprime, gradprime, logpprime = leapfrog(theta, r, grad, v * epsilon, f)
            jointprime = logpprime - 0.5 * np.dot(rprime, rprime.T)
            # Is the simulation wildly inaccurate?
            sprime = jointprime - joint0 > -1000
            # Set the return values---minus=plus for all things here, since the
            # "tree" is of depth 0.
            thetaminus = thetaprime[:]
            thetaplus = thetaprime[:]
            rminus = rprime[:]
            rplus = rprime[:]
            gradminus = gradprime[:]
            gradplus = gradprime[:]
            logptree = jointprime - joint0
            #logptree = logpprime
            # Compute the acceptance probability.
            alphaprime = min(1., np.exp(jointprime - joint0))
            #alphaprime = min(1., np.exp(logpprime - 0.5 * np.dot(rprime, rprime.T) - joint0))
            nalphaprime = 1
        else:
            # Recursion: Implicitly build the height j-1 left and right subtrees.
            thetaminus, rminus, gradminus, thetaplus, rplus, gradplus, thetaprime, gradprime, logpprime, sprime, alphaprime, nalphaprime, logptree = build_tree(theta, r, grad, v, j - 1, epsilon, f, joint0)
            # No need to keep going if the stopping criteria were met in the first subtree.
            if sprime:
                if v == -1:
                    thetaminus, rminus, gradminus, _, _, _, thetaprime2, gradprime2, logpprime2, sprime2, alphaprime2, nalphaprime2, logptree2 = build_tree(thetaminus, rminus, gradminus, v, j - 1, epsilon, f, joint0)
                else:
                    _, _, _, thetaplus, rplus, gradplus, thetaprime2, gradprime2, logpprime2, sprime2, alphaprime2, nalphaprime2, logptree2 = build_tree(thetaplus, rplus, gradplus, v, j - 1, epsilon, f, joint0)
                # Conpute total probability of this trajectory
                logptot = np.logaddexp(logptree, logptree2)
                # Choose which subtree to propagate a sample up from.
                if np.log(np.random.uniform()) < logptree2 - logptot:
                    thetaprime = thetaprime2[:]
                    gradprime = gradprime2[:]
                    logpprime = logpprime2
                logptree = logptot
                # Update the stopping criterion.
                sprime = sprime and sprime2 and stop_criterion(thetaminus, thetaplus, rminus, rplus)
                # Update the acceptance probability statistics.
                alphaprime = alphaprime + alphaprime2
                nalphaprime = nalphaprime + nalphaprime2

        return thetaminus, rminus, gradminus, thetaplus, rplus, gradplus, thetaprime, gradprime, logpprime, sprime, alphaprime, nalphaprime, logptree

def tree_sample(theta, logp, r0, grad, epsilon, f, joint, maxheight=np.inf):
    pass



from __future__ import division, print_function
import sys
import os
import dill
import time
import logging
import bisect
import numpy as np
from numpy import logaddexp
from numpy import inf
from scipy.stats import kstest
from math import isnan
from . import nest2pos
from .nest2pos import logsubexp
from operator import attrgetter
from .raynest import CheckPoint
from .utils import auto_garbage_collect
import copy

import ray
from tqdm import tqdm
logger = logging.getLogger('raynest.NestedSampling')

class KeyOrderedList(list):
    """
    List object that is ordered according to a key

    Parameters
    ----------
    iterable : array_like
        Initial input used to intialise the list
    key : function, optional
        Key to use to sort the list, by defaul it is sorted by its
        values.
    """
    def __init__(self, iterable, key=lambda x: x):
        iterable = sorted(iterable, key=key)
        super(KeyOrderedList, self).__init__(iterable)

        self._key = key
        self._keys = [self._key(v) for v in iterable]

    def search(self, item):
        """
        Find the location of a new entry
        """
        return bisect.bisect(self._keys, self._key(item))

    def add(self, item):
        """
        Update the ordered list with a single item and return the index
        """
        index = self.search(item)
        self.insert(index, item)
        self._keys.insert(index, self._key(item))
        return index


class OrderedLivePoints(KeyOrderedList):
    """
    Object that contains live points ordered by increasing log-likelihood. Requires
    the log-likelihood to be pre-computed.

    Assumes the log-likelihood is accesible as an attribute of each live point.

    Parameters
    ----------
    live_points : array_like
        Initial live points
    """
    def __init__(self, live_points):
        super(OrderedLivePoints, self).__init__(live_points, key=lambda x: x.logL)

    def insert_live_point(self, live_point):
        """
        Insert a live point and return the index of the new point
        """
        return self.add(live_point)

    def remove_n_worst_points(self, n):
        """
        Remove the n worst live points
        """
        del self[:n]
        del self._keys[:n]


class _NSintegralState(object):
    """
    Stores the state of the nested sampling integrator
    """
    def __init__(self, nlive):
        self.nlive = nlive
        self.reset()
        loggername = 'raynest.NestedSampling._NSintegralState'
        self.logger = logging.getLogger(loggername)

    def reset(self):
        """
        Reset the sampler to its initial state at logZ = -infinity
        """
        self.iteration = 0
        self.logZ = -inf
        self.oldZ = -inf
        self.logw = 0
        self.info = 0
        # Start with a dummy sample enclosing the whole prior
        self.logLs = [-inf]  # Likelihoods sampled
        self.log_vols = [0.0]  # Volumes enclosed by contours

    def increment(self, logL, nlive=None, nreplace=1):
        """
        Increment the state of the evidence integrator
        Simply uses rectangle rule for initial estimate
        see: https://www.cell.com/biophysj/pdf/S0006-3495(12)00055-0.pdf
        for parallel implementation
        """
        if(logL<=self.logLs[-1]):
            self.logger.warning('NS integrator received non-monotonic logL. {0:.5f} -> {1:.5f}'.format(self.logLs[-1], logL))
        if nlive is None:
            nlive = self.nlive
        oldZ = self.logZ
        logt=-nreplace/nlive
        Wt = self.logw + logL + logsubexp(0,logt)
        self.logZ = logaddexp(self.logZ,Wt)
        # Update information estimate
        if np.isfinite(oldZ) and np.isfinite(self.logZ) and np.isfinite(logL):
            self.info = np.exp(Wt - self.logZ)*logL + np.exp(oldZ - self.logZ)*(self.info + oldZ) - self.logZ
            if isnan(self.info):
                self.info = 0

        # Update history
        self.logw += logt
        self.iteration += 1
        self.logLs.append(logL)
        self.log_vols.append(self.logw)

    def finalise(self):
        """
        Compute the final evidence with more accurate integrator
        Call at end of sampling run to refine estimate
        """
        from scipy import integrate
        # Trapezoidal rule
        self.logZ = nest2pos.log_integrate_log_trap(np.array(self.logLs), np.array(self.log_vols))
        return self.logZ

    def plot(self,filename):
        """
        Plot the logX vs logL
        """
        import matplotlib as mpl
        mpl.use('Agg')
        from matplotlib import pyplot as plt
        plt.figure()
        plt.plot(self.log_vols,self.logLs)
        plt.title('{0} iterations. logZ={1:.2f} H={2:.2f} bits'.format(self.iteration,self.logZ,self.info*np.log2(np.e)))
        plt.grid(which='both')
        plt.xlabel('log prior_volume')
        plt.ylabel('log likelihood')
        plt.xlim([self.log_vols[-1],self.log_vols[0]])
        plt.savefig(filename)
        self.logger.info('Saved nested sampling plot as {0}'.format(filename))

    def __getstate__(self):
        """Remove the unpicklable entries."""
        state = self.__dict__.copy()
        del state['logger']
        return state

    def __setstate__(self, state):
        if 'logger' not in state:
            loggername = "raynest.NestedSampling._NSintegralState"
            state['logger'] = logging.getLogger(loggername)
        self.__dict__ = state

class NestedSampler(object):
    """
    Nested Sampler class.
    Initialisation arguments:

    model: :obj:`raynest.Model` user defined model

    periodic_checkpoint_interval: float
        time interval in seconds between periodic checkpoints

    nthreads: int
        number of parallel sampling threads

    nlive: int
        number of live points to be used for the integration
        Default: 1024

    output: string
        folder where the output will be stored
        Default: None

    verbose: int
        0: Nothing
        1: display information on screen
        2: (1) + diagnostic plots
        Default: 1

    seed: int
        seed for the initialisation of the pseudorandom chain
        Default: 1234

    prior_sampling: boolean
        produce Nlive samples from the prior.
        Default: False

    stopping: float
        Stop when remaining samples wouldn't change logZ estimate by this much.
        Deafult: 0.1

    n_periodic_checkpoint: int
        **deprecated**
        This parameter should not be used, it should be set by the manager instead.
        checkpoint the sampler every n_periodic_checkpoint iterations
        Default: None (disabled)

    resume_file: string
        file where the checkpoints will be stored
        Default: None

    state: dict
        dictionary holding the checkpointed state
        Default: None
    """

    def __init__(self,
                 model,
                 periodic_checkpoint_interval = np.inf,
                 nthreads       = None,
                 nlive          = 1024,
                 output         = None,
                 verbose        = 1,
                 rng            = None,
                 prior_sampling = False,
                 stopping       = 0.1,
                 n_periodic_checkpoint = None,
                 position       = 0,
                 resume_file    = None,
                 state          = None
                 ):
        """
        Initialise all necessary arguments and
        variables for the algorithm
        """
        loggername          = 'raynest.NestedSampling.NestedSampler'
        self.logger         = logging.getLogger(loggername)
        self.logger.addHandler(logging.StreamHandler())
        self.model          = model

        if state is None:
            if rng == None:
                self.rng = np.random.default_rng()
            else:
                self.rng            = rng
            self.position       = position
            self.periodic_checkpoint_interval = periodic_checkpoint_interval
            self.nthreads       = nthreads
            self.prior_sampling = prior_sampling
            self.verbose        = verbose
            self.acceptance     = 1.0
            self.accepted       = 0
            self.rejected       = 1
            self.nlive          = nlive
            self.live_points    = None
            self.last_checkpoint_time = time.time()
            self.tolerance      = stopping
            self.condition      = np.inf
            self.worst          = 0
            self.logLmin        = -np.inf
            self.logLmax        = -np.inf
            self.iteration      = 0
            self.prior_samples  = []
            self.nested_samples = []
            self.rolling_p      = []
            self.logZ           = None
            self.resume_file    = resume_file
            sys.stdout.flush()
            self.output_folder  = output
            self.output_file,self.evidence_file = self.setup_output(output)
            header              = open(os.path.join(output,'header.txt'),'w')
            header.write('\t'.join(self.model.names))
            header.write('\tlogL\n')
            header.close()
            self.initialise_live_points()
            self.initialised    = False
        else:
            for k,v in state.items():
                setattr(self,k,v)
            # silly workaround for extra live points if caught in the middle of
            # updating them
            if self.nlive != len(self.live_points._list):
                self.live_points._list.pop(-1)

    def initialise_live_points(self):

        l = []

        with tqdm(total=self.nlive, disable = not self.verbose, desc='raynest: populate samplers', position=self.position) as pbar:
            for i in range(self.nlive):

                p = self.model.new_point(rng = self.rng)
                p.logP = self.model.log_prior(p)
                p.logL = self.model.log_likelihood(p)
                l.append(p)
                pbar.update()
        
        self.live_points = LivePoints(l, self.rng, n_replace = self.nthreads)

    def setup_output(self, output):
        """
        Set up the output folder

        -----------
        Parameters:
        output: string
            folder where the results will be stored
        -----------
        Returns:
            output_file, evidence_file, resume_file: tuple
                output_file:   file where the nested samples will be written
                evidence_file: file where the evidence will be written
                resume_file:   file used for checkpointing the algorithm
        """
        chain_filename = "chain_"+str(self.nlive)+"_"+str(self.position)+".txt"
        output_file   = os.path.join(output,chain_filename)
        evidence_file = os.path.join(output,chain_filename+"_evidence_"+str(self.position)+".txt")

        return output_file, evidence_file


    def write_chain_to_file(self):
        """
        Outputs a `raynest.parameter.LivePoint` to the
        output_file
        """
        with open(self.output_file,"w") as f:
            f.write('{0:s}\n'.format(self.model.header().rstrip()))
            for ns in self.nested_samples:
                f.write('{0:s}\n'.format(self.model.strsample(ns).rstrip()))

    def write_evidence_to_file(self):
        """
        Write the evidence logZ and maximum likelihood to the evidence_file
        """
        with open(self.evidence_file,"w") as f:

            f.write('#logZ\tlogLmax\tH\n')
            f.write('{0:.5f} {1:.5f} {2:.2f}\n'.format(self.live_points.get_logZ(),
                                                       self.live_points.get_logLmax(),
                                                       self.live_points.get_info()))

    def consume_sample(self, pool):
        """
        requests the workers to update the live points
        and updates the evidence logZ
        """
        # Get the worst live points
        self.worst   = self.live_points.get_worst(self.nthreads)
        # get the necessary statistics from the LivePoint actor
        self.logLmin, self.logLmax, logZ, info = self.live_points.get_logLs_logZ_info()

        if self.verbose:
            logLtmp = self.live_points.get_worst_logLs()

        self.condition = logaddexp(logZ,self.logLmax - self.iteration/(float(self.nlive))) - logZ
        
        # set up the ensemble statistics
        if (self.iteration % (self.nlive//10)) < self.nthreads:
            lpp = LivePoints(self.live_points._list, self.rng, n_replace = self.nthreads)
            lp = ray.put(lpp)
            for s in pool.map(lambda a, v: a.set_ensemble.remote([lp]), range(self.nthreads)):
                pass
            
        starting_points = self.live_points.sample(self.nthreads)

        for v in starting_points:
            pool.submit(lambda a, v: a.produce_sample.remote(v, self.logLmin), v)

        i = 0
        while pool.has_next():

            acceptance, sub_acceptance, self.jumps, proposed = pool.get_next()

            if proposed.logL > self.logLmin:
                # replace worst point with new one
                self.live_points.insert(i%self.nthreads,proposed.copy())
                self.accepted  += 1
                self.iteration += 1

                if self.verbose:
                    self.logger.info("{0:d}: n:{1:4d} NS_acc:{2:.3f} S{3:02d}_acc:{4:.3f} sub_acc:{5:.3f} H: {6:.2f} logL {7:.5f} --> {8:.5f} dZ: {9:.3f} logZ: {10:.3f} logLmax: {11:.2f}"\
                        .format(self.iteration, self.jumps, self.acceptance, i%self.nthreads, acceptance, sub_acceptance, info,\
                        logLtmp[i%self.nthreads], proposed.logL, self.condition, logZ, self.logLmax))

                i += 1

            else:
                self.rejected += 1
                p = pool.submit(lambda a, v: a.produce_sample.remote(v, self.logLmin), starting_points[i%self.nthreads])

            self.acceptance = float(self.accepted)/float(self.accepted + self.rejected)

        self.live_points.remove_n_worst_points(self.nthreads)
        self.live_points.update_mean_covariance()

    def reset(self, pool):
        """
        Initialise the pool of `raynest.parameter.LivePoint` by
        sampling them from the `raynest.model.log_prior` distribution
        """
        # set up  the ensemble statistics
        lpp = LivePoints(self.live_points._list, self.rng, n_replace = self.nthreads)
        lp = ray.put(lpp)
        
        for s in pool.map(lambda a, v: a.set_ensemble.remote([lp]), range(self.nthreads)):
            pass
        
        # send all live points to the samplers for start
        for i in range(self.nlive):
            pool.submit(lambda a, v: a.produce_sample.remote(v, -np.inf), self.live_points.get(i))

        i = 0

        with tqdm(total=self.nlive, disable= not self.verbose, desc='raynest: sampling prior', position=self.position) as pbar:
            while pool.has_next():
                acceptance,sub_acceptance,self.jumps,x = pool.get_next()
                if np.isnan(x.logL):
                    self.logger.warning("Likelihood function returned NaN for params "+str(x))
                    self.logger.warning("You may want to check your likelihood function")
                if np.isfinite(x.logP) and np.isfinite(x.logL):
                    self.live_points.set(i,x)
                    self.prior_samples.append(x)
                    i += 1
                    pbar.update()
                else:
                    pool.submit(lambda a, v: a.produce_sample.remote(v, -np.inf), self.live_points.get(i))

        self.live_points.update_mean_covariance()
        self.live_points.set_ordered_list()

        if self.verbose:
            sys.stderr.write("\n")
            sys.stderr.flush()
        self.initialised=True

    def nested_sampling_loop(self, pool):
        """
        main nested sampling loop
        """
        if not self.initialised:
            self.reset(pool)
        else:
            self.logger.info("Nested Sampling process {0!s}, restoring samplers".format(os.getpid()))
            lp_ref = ray.put(self.live_points)
            for s in pool.map(lambda a, v: a.set_ensemble.remote([lp_ref]), range(self.nthreads)):
                pass

        if self.prior_sampling:
            self.logZ, self.nested_samples = self.live_points.finalise()
            self.write_chain_to_file()
            self.write_evidence_to_file()
            self.logLmin = np.inf
            self.logLmax = np.inf
            self.logger.info("Nested Sampling process {0!s}, exiting".format(os.getpid()))
            return 0

        try:
            while self.condition > self.tolerance:
                self.consume_sample(pool)
                if time.time() - self.last_checkpoint_time > self.periodic_checkpoint_interval:
                    self.save()
                    self.last_checkpoint_time = time.time()

                if (self.iteration % self.nlive) < self.nthreads:
                    self.check_insertion_indices()
                auto_garbage_collect()

        except CheckPoint:
            self.checkpoint()
            sys.exit(130)

        # Refine evidence estimate
        self.logZ, self.nested_samples = self.live_points.finalise()
        self.info = self.live_points.get_info()
        # output the chain and evidence
        self.write_chain_to_file()
        self.write_evidence_to_file()
        if self.verbose:
            self.logger.critical('Final evidence: {0:0.2f}'.format(self.logZ))
            self.logger.critical('Information: {0:.2f}'.format(self.info))

        # Some diagnostics
        if self.verbose > 1 :
            self.live_points.plot(os.path.join(self.output_folder,'logXlogL.png'))

        return 0

    def check_insertion_indices(self, rolling=True, filename=None):
        """
        Checking the distibution of the insertion indices either during
        the nested sampling run (rolling=True) or for the whole run
        (rolling=False).
        """
        if not self.live_points.get_insertion_indices():
            return
        if rolling:
            indices = self.live_points.get_insertion_indices()[-self.nlive:]
        else:
            indices = self.live_points.get_insertion_indices()

        D, p = kstest(indices, 'uniform', args=(0, 1))
        if rolling:
            if self.verbose:
                self.logger.info('Rolling KS test: D={0:.3}, p-value={1:.3}'.format(D, p))
            self.rolling_p.append(p)
        else:
            if self.verbose:
                self.logger.info('Final KS test: D={0:.3}, p-value={1:.3}'.format(D, p))

        if filename is not None:
            np.savetxt(os.path.join(
                self.output_folder, filename),
                self.live_points.get_insertion_indices(),
                newline='\n',delimiter=' ')

    def get_output_folder(self):
        return self.output_folder

    def get_nested_samples(self):
        return self.nested_samples

    def get_prior_samples(self):
        return self.prior_samples

    def get_live_points(self):
        return self.live_points

    def get_logZ(self):
        return self.logZ

    def get_nlive(self):
        return self.nlive

    def get_information(self):
        return self.info

    def save(self):
        state = self.__dict__.copy()
        self.logger.critical('Saving Nested Sampling state in ' + str(self.resume_file))
        del state['model']
        del state['logger']

        with open(self.resume_file,'wb') as f:
            dill.dump(state,f)

class LivePoints:
    """
    class holding the live points pool

    parameters
    ==========

    l: list
        list of live points generated from the `raynest.model` new_point method

    n_replace: int
        number of live points to be replaced at each step. must coincide with the number of
        sampling threads

    verbose: int
        level of verbosity
    """
    def __init__(self, l, rng, n_replace = 1, verbose = 2):
        self.rng                  = rng
        self.n_replace            = n_replace
        self._list                = l
        self.n                    = len(l)
        self.dim                  = self._list[0].dimension
        self.mean                 = None
        self.covariance           = None
        self.eigen_values         = None
        self.eigen_vectors        = None
        self.likelihood_gradient  = None
        self.worst                = None
        self.logLmax              = -np.inf
        self.logLmin              = -np.inf
        self.state                = _NSintegralState(self.n)
        self.logger               = logging.getLogger('raynest')
        self.nested_samples       = []
        self.insertion_indices    = []
        self.update_mean_covariance()

    def set_ordered_list(self):
        """
        initialise the list of live points as a ordered list
        """
        self._list = OrderedLivePoints(self._list)

    def get(self, i):
        """
        return the i-th element
        """
        return self._list[i]

    def get_mean_covariance(self):
        """
        return the live points sample mean and covariance
        """
        return self.get_mean(), self.get_covariance()

    def get_covariance(self):
        """
        return the live points sample covariance
        """
        return self.covariance

    def get_mean(self):
        """
        return the live points sample mean
        """
        return self.mean

    def get_eigen_quantities(self):
        """
        return the live points sample covariance eigen values and eigen vectors
        """
        return self.eigen_values, self.eigen_vectors

    def get_length(self):
        """
        return the number of live points
        """
        return self.n

    def get_dimension(self):
        """
        return the dimension of each live point
        """
        return self.dim

    def set(self,i, val):
        """
        set the i-th live point to val
        """
        self._list[i] = val

    def insert(self, i, val):
        """
        insert val in the ordered list
        """
        index = self._list.insert_live_point(val)
        self.insertion_indices.append((index - self.n_replace) / (self.n - i - 1))

    def to_list(self):
        """
        return the list of live points
        """
        return self._list

    def update_mean_covariance(self):
        """
        Recompute mean and covariance matrix
        of the ensemble of Live points
        """
        cov_array = np.zeros((self.dim,self.n))
        if self.dim == 1:
            name=self._list[0].names[0]
            self.covariance = np.atleast_2d(np.var([self._list[j][name] for j in range(self.n)]))
            self.mean       = np.atleast_1d(np.mean([self._list[j][name] for j in range(self.n)]))
        else:
            for i,name in enumerate(self._list[0].names):
                for j in range(self.n): cov_array[i,j] = self._list[j][name]
            self.covariance = np.cov(cov_array)
            self.mean       = np.mean(cov_array,axis=1)
        self.eigen_values, self.eigen_vectors = np.linalg.eigh(self.covariance)

    def get_as_array(self):
        """
        return the live points as a stacked (n,d) numpy array
        """
        n = self.n+len(self.nested_samples)
        as_array = np.zeros((self.dim,n))
        for i,name in enumerate(self._list[0].names):
            for j in range(self.n): as_array[i,j] = self._list[j][name]
            for j in range(len(self.nested_samples)): as_array[i,j+self.n] = self.nested_samples[j][name]

        return as_array.T # as an array (N,D)

    def sample(self, n):
        """
        randomly sample n live points
        """
        if self.worst == None:
            return [self._list[i] for i in self.rng.choice(self.n, n, replace=False)]
        else:
            return [self._list[i] for i in self.rng.choice(self.n-n, n, replace=False)]

    def sample_around(self, q, n):
        """
        sample n live points closest to q
        """
        distance = lambda list_value : np.sum(list_value.values - q.values)
        closest_value = min(self._list, key=distance)
        idx = self._list.index(closest_value)
        return [self.get(i) for i in range(n//2-idx,n//2+idx)]

    def get_logLmax(self):
        """
        return the maximum likelihood value
        """
        return self.logLmax

    def get_logLmin(self):
        """
        return the minimum likelihood value
        """
        return self.logLmin

    def get_worst(self, n):
        """
        selects the lowest likelihood N live points
        and updates the integral state
        """
        self.logLmin = self._list[n-1].logL
        self.logLmax = self._list[-1].logL
        self.worst = self._list[:n]
        self.nested_samples.extend(self.worst)
        self.state.increment(self.worst[n-1].logL, nreplace=self.n_replace)
        return self.worst

    def remove_n_worst_points(self,n):
        """
        remove the set of worst n live points
        """
        self._list.remove_n_worst_points(n)

    def get_insertion_indices(self):
        """
        return the insertion indeces
        """
        return self.insertion_indices

    def get_info(self):
        """
        return the information
        """
        return self.state.info

    def get_logZ(self):
        """
        return the log evidence
        """
        return self.state.logZ

    def get_logLs_logZ_info(self):
        return self.get_logLmin(), self.get_logLmax(), self.get_logZ(), self.get_info()

    def get_worst_logLs(self):
        return np.array([w.logL for w in self.worst])

    def finalise(self):
        # final adjustments
        self._list.sort(key=attrgetter('logL'))
        for i,p in enumerate(self._list):
            self.state.increment(p.logL, nlive=self.n-i, nreplace = 1)
            self.nested_samples.append(p)

        # Refine evidence estimate
        self.logZ = self.state.finalise()
        return self.state.logZ, self.nested_samples

    def plot(self,filename):
        self.state.plot(filename)
        return 0

    def _get_integral_state(self):
        return self.state

    def _set_internal_state(self, state):
        self.state = state



# coding: utf-8

import numpy as np
import os
import sys
import signal
import logging
import dill

from .utils import LogFile

# module logger takes name according to its path
LOGGER = logging.getLogger('raynest.raynest')
import ray
from ray.util import ActorPool
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy

class CheckPoint(Exception):
    pass

def sighandler(signal, frame):
    LOGGER.critical("Handling signal {}".format(signal))
    raise CheckPoint()

class raynest(object):
    """
    Class to control raynest sampler
    cp = raynest(usermodel,nlive=100,output='./',verbose=0,seed=None,maxmcmc=100,nthreads=None,balanced_sampling = True)

    Input variables:
    =====

    usermodel: :obj:`raynest.Model`
        a user-defined model to analyse

    nlive: `int`
        Number of live points (100)

    output : `str`
        output directory (./)

    verbose: `int`
        Verbosity:
            0: no display of information on screen, save the NS chain and evidence
            1: 0 + display progress on screen
            2: 1 + display diagnostics (ACL), save the posterior samples and trace plots and posterior plots
            3: 2 + save chains from individual samplers
            default: 0

    seed: `int`
        random seed (default: 1234)

    maxmcmc: `int`
        maximum MCMC points for MHS sampling chains (5000)

    nnest: `int`
        number of parallel nested samplers.
        Specifying nnest > 1 will run multiple parallel nested samplers, and combine the
        results at the end to produce a single evidence and set of posterior samples.
        Default: 1

    nensemble: `int`
        total number of ensemble samplers. each nested sampler instance is assigned nensemble/nnest. Default: 1

    nhamiltonian: `int`
        ntotal number of hamitonian samplers. each nested sampler instance is assigned nhamiltonian/nnest. Default: 0

    nslice: `int`
        ntotal number of slice samplers. each nested sampler instance is assigned nslice/nnest. Default: 0

    resume: `boolean`
        determines whether raynest will resume a run or run from scratch. Default: False.

    proposals: `dict`
        dictionary of lists with custom jump proposals.
        key 'mhs' for the Metropolis-Hastings sampler,
        'hmc' for the Hamiltonian Monte-Carlo sampler,
        'sli' for the slice sampler.
        Default: None
        By default, the sampler will use the  :obj:`raynest.proposal.DefaultProposalCycle`
        for the mhs sampler; :obj:`raynest.proposal.HamiltonianProposalCycle` for hmc and
        :obj:`raynest.proposal.EnsembleSliceProposalCycle` for the sli sampler.

    prior_sampling: `boolean`
        Default: False
        If true, generates samples from the prior then terminates. Adjust `nlive` to control the number
        of samples requested.

    periodic_checkpoint_interval: `float`
        checkpoing the sampler every periodic_checkpoint_interval seconds
        Default: None (disabled)

    object_store_memory: `int`
        amount of memory in bytes reserved for ray object store
        Default: 1e9 (1GB)

    """
    def __init__(self,
                 usermodel,
                 nlive        = 100,
                 output       = './',
                 verbose      = 0,
                 seed         = None,
                 maxmcmc      = 5000,
                 nnest        = 1,
                 nensemble    = 0,
                 nhamiltonian = 0,
                 nslice       = 0,
                 resume       = False,
                 proposals    = None,
                 periodic_checkpoint_interval=None,
                 prior_sampling = False,
                 object_store_memory=10**9
                 ):

        self.verbose   = verbose
        self.logger    = logging.getLogger('raynest.raynest.raynest')
        self.nsamplers = nensemble+nhamiltonian+nslice
        self.nnest     = nnest
        
        assert self.nsamplers > 0, "no sampler processes requested!"
        import psutil
        self.max_threads = psutil.cpu_count()

        if self.nsamplers%self.nnest != 0:
            self.logger.warning("Error! Number of samplers not balanced")
            self.logger.warning("to the number of nested samplers! Exiting.")
            exit(-1)

        self.nthreads = self.nsamplers+self.nnest

        if self.nthreads > self.max_threads:
            self.logger.warning("More cpus than available are being requested!")
            self.logger.warning("This might result in excessive overhead")

        self.ns_pool = []
        self.pool    = []

        try:
            ray.init(address='auto')
            self.existing_cluster = True
        except:
            ray.init(num_cpus=self.nthreads,
                     ignore_reinit_error=True,
                     object_store_memory=object_store_memory)
            self.existing_cluster = False
                     
        assert ray.is_initialized() == True
        output = os.path.join(output, '')
        checkpoint_folder = os.path.join(output,'checkpoints')
        os.makedirs(output, exist_ok=True)
        os.makedirs(checkpoint_folder, exist_ok=True)
        
        # Import placement group APIs.
        from ray.util.placement_group import placement_group, placement_group_table, remove_placement_group

        # set up the resume files
        self.resume_file = []
        # The LogFile context manager ensures everything within is logged to
        # 'raynest.log' but the file handler is safely closed once the run is
        # finished.
        self.log_file = LogFile(os.path.join(output, 'raynest.log'),
                                verbose=self.verbose)

        with self.log_file:

            if self.verbose:
                self.logger.info('Running with {0} parallel threads'.format(self.nthreads))
                self.logger.info('Nested samplers: {0}'.format(nnest))
                self.logger.info('Ensemble samplers: {0}'.format(nensemble))
                self.logger.info('Slice samplers: {0}'.format(nslice))
                self.logger.info('Hamiltonian samplers: {0}'.format(nhamiltonian))
                self.logger.info('ray object store size: {0} GB'.format(object_store_memory/1e9))

            if periodic_checkpoint_interval is None:
                self.periodic_checkpoint_interval = np.inf
            else:
                self.periodic_checkpoint_interval = periodic_checkpoint_interval

            from .sampler import HamiltonianMonteCarloSampler, MetropolisHastingsSampler, SliceSampler, SamplersCycle
            from .NestedSampling import NestedSampler
            from .proposal import DefaultProposalCycle, HamiltonianProposalCycle, EnsembleSliceProposalCycle
            if proposals is None:
                proposals = dict(mhs=DefaultProposalCycle,
                                 hmc=HamiltonianProposalCycle,
                                 sli=EnsembleSliceProposalCycle)
            elif type(proposals) == list:
                proposals = dict(mhs=proposals[0],
                                 hmc=proposals[1],
                                 sli=proposals[2])
            self.nlive    = nlive
            self.output   = output
            self.results  = {}
            self.prior_sampling = prior_sampling
            self.user     = usermodel
            self.resume = resume
            
            if seed is not None:
                self.seed = iter([np.random.SeedSequence(seed+i) for i in range(self.nthreads)])
            else:
                self.logger.warning('seed = None was passed. The results will not be reproducible. \
                                     Set seed = int to have a reproducible sequence.')
                self.seed = seed
            
            rng = None
            for j in range(self.nnest):
                
                if self.seed is not None:
                    s0 = next(self.seed)
                    rng = np.random.default_rng(s0)

                pg = placement_group([{"CPU": 1+self.nsamplers//self.nnest}],strategy="PACK")
                ray.get(pg.ready())
                
                self.resume_file.append(os.path.join(checkpoint_folder, "nested_sampler_resume_{}.pkl".format(j)))
                
                if not os.path.exists(self.resume_file[j]) or resume == False:
                    
                    self.ns_pool.append(ray.remote(NestedSampler).options(scheduling_strategy=PlacementGroupSchedulingStrategy(
        placement_group=pg)).remote(
                                self.user,
                                nthreads       = self.nsamplers,
                                nlive          = nlive,
                                output         = output,
                                verbose        = self.verbose,
                                rng            = rng,
                                prior_sampling = self.prior_sampling,
                                periodic_checkpoint_interval = self.periodic_checkpoint_interval,
                                resume_file    = self.resume_file[j],
                                position = j))
                else:
                    state = self.load_nested_sampler_state(self.resume_file[j])
                    ns = ray.remote(NestedSampler).options(scheduling_strategy=PlacementGroupSchedulingStrategy(
        placement_group=pg)).remote(
                                self.user,
                                nthreads       = self.nsamplers,
                                nlive          = nlive,
                                output         = output,
                                verbose        = self.verbose,
                                prior_sampling = self.prior_sampling,
                                periodic_checkpoint_interval = self.periodic_checkpoint_interval,
                                resume_file    = self.resume_file[j],
                                position = j,
                                state    = state)

                    self.ns_pool.append(ns)
                
                samplers = []

                # instantiate the sampler class
                for i in range(nensemble//self.nnest):
                    
                    if self.seed is not None:
                        rng = np.random.default_rng(next(self.seed))
                    
                    S = MetropolisHastingsSampler.options(scheduling_strategy=PlacementGroupSchedulingStrategy(
        placement_group=pg)).remote(self.user,
                                          maxmcmc,
                                          rng         = rng,
                                          verbose     = self.verbose,
                                          proposal    = proposals['mhs']
                                          )
                    samplers.append(S)
                
                for i in range(nhamiltonian//self.nnest):
                    
                    if self.seed is not None:
                        rng = np.random.default_rng(next(self.seed))
                        
                    S = HamiltonianMonteCarloSampler.options(scheduling_strategy=PlacementGroupSchedulingStrategy(
        placement_group=pg)).remote(self.user,
                                          maxmcmc,
                                          rng         = rng,
                                          verbose     = self.verbose,
                                          proposal    = proposals['hmc']
                                          )
                    samplers.append(S)
                    
                for i in range(nslice//self.nnest):

                    if self.seed is not None:
                        rng = np.random.default_rng(next(self.seed))
                        
                    S = SliceSampler.options(scheduling_strategy=PlacementGroupSchedulingStrategy(
        placement_group=pg)).remote(self.user,
                                          maxmcmc,
                                          rng         = rng,
                                          verbose     = self.verbose,
                                          proposal    = proposals['sli']
                                          )
                    samplers.append(S)
                
                self.pool.append(ActorPool(samplers))
                self.results['run_{}'.format(j)] = {}
                
                    
            if self.seed is not None:
                self.results['seed'] = self.seed

            self.results['combined'] = {}

    def run(self,
            postprocess = True,
            plot        = True,
            corner      = False):
        """
        Run the sampler
        ===============
        kwargs:
        postprocess: generates and saves posterior samples. default = True
        
        plot: generates summary plots. default = True
        
        corner: generates a summary corner plot. Warning! in large dimansional spaces this might take a while. default = False
        """
        with self.log_file:
            if self.resume:
                signal.signal(signal.SIGTERM, sighandler)
                signal.signal(signal.SIGALRM, sighandler)
                signal.signal(signal.SIGQUIT, sighandler)
                signal.signal(signal.SIGINT, sighandler)
                signal.signal(signal.SIGUSR1, sighandler)
                signal.signal(signal.SIGUSR2, sighandler)
            
            try:
                unfinished = [self.ns_pool[v].nested_sampling_loop.remote(self.pool[v]) for v in  range(self.nnest)]
                while len(unfinished) > 0:
                    finished, unfinished = ray.wait(unfinished)
                    ray.get(finished)

            except CheckPoint:
                self.checkpoint()
                for p in self.ns_pool:
                    ray.kill(p)
                if self.existing_cluster is False:
                    ray.shutdown()
                assert ray.is_initialized() == False
                self.logger.critical("ray correctly shut down. exiting")
                sys.exit(130)

            if self.verbose > 0:
                filename = 'raynest.h5'
            else:
                filename = None
            
            self.postprocess(filename=filename)

            if self.verbose >= 2:
                self.logger.critical("Checking insertion indeces")
                ray.get([self.ns_pool[v].check_insertion_indices.remote(
                                    rolling=False,
                                    filename='insertion_indices_{}.dat'.format(v))
                                    for v in range(self.nnest)])
 
                self.logger.critical("Saving plots in {0}".format(self.output))
                self.plot(corner = False)

            try:
                for f in self.resume_file:
                    self.logger.info("Removing checkpoint file {}".format(f))
                    os.remove(f)
            except OSError:
                pass
            
            if self.existing_cluster is False:
                ray.shutdown()
                assert ray.is_initialized() == False
                self.logger.critical("ray correctly shut down. exiting")

    def postprocess(self, filename=None):
        """
        post-processes the results of parallel runs
        if filename is not None, then outputs the results in an
        hdf5 file
        """
        import numpy.lib.recfunctions as rfn
        from .nest2pos import draw_posterior_many

        ns = ray.get([self.ns_pool[v].get_nested_samples.remote() for v in range(self.nnest)])
        ps = ray.get([self.ns_pool[v].get_prior_samples.remote() for v in range(self.nnest)])
        info = ray.get([self.ns_pool[v].get_information.remote() for v in range(self.nnest)])

        for i,l in enumerate(ps):

            self.results['run_{}'.format(i)]['prior_samples'] = rfn.stack_arrays([s.asnparray()
                               for s in l] ,usemask=False)

        for i,l in enumerate(ns):
            self.results['run_{}'.format(i)]['nested_samples'] = rfn.stack_arrays([s.asnparray()
                               for s in l] ,usemask=False)

        for i in range(self.nnest):
            p, logZ  = draw_posterior_many([self.results['run_{}'.format(i)]['nested_samples']],
                                           [self.nlive],verbose=self.verbose)

            self.results['run_{}'.format(i)]['posterior_samples'] = p
            self.results['run_{}'.format(i)]['logZ'] = logZ
            self.results['run_{}'.format(i)]['information'] = info[i]
            self.results['run_{}'.format(i)]['logZ_error'] = np.sqrt(info[i]/self.nlive)

        p, logZ  = draw_posterior_many([self.results['run_{}'.format(i)]['nested_samples'] for i in range(self.nnest)],[self.nlive]*self.nnest,verbose=self.verbose,rng=np.random.default_rng(666))

        self.results['combined']['prior_samples'] = rfn.stack_arrays([self.results['run_{}'.format(i)]['prior_samples'] for i in range(self.nnest)],usemask=False)
        self.results['combined']['nested_samples'] = rfn.stack_arrays([self.results['run_{}'.format(i)]['nested_samples'] for i in range(self.nnest)],usemask=False)

        self.results['combined']['nested_samples'].sort(order='logL')
        self.results['combined']['posterior_samples'] = p
        self.results['combined']['logZ'] = logZ
        self.results['combined']['information'] = np.average(info)

        evds = [self.results['run_{}'.format(i)]['logZ'] for i in range(self.nnest)]
        e2   = (np.average(info)/self.nlive+np.var(evds))/self.nnest
        self.results['combined']['logZ_error'] = np.sqrt(e2)

        if filename != None:
            import os
            import h5py
            self.logger.critical("Saving output in {0}".format(self.output))
            h = h5py.File(os.path.join(self.output,filename),'w')
            for k in self.results.keys():
                if k != 'seed':
                    grp = h.create_group(k)
                    for d in self.results[k].keys():
                        grp.create_dataset(d, data = self.results[k][d], dtype=self.results[k][d].dtype)
            h.close()

    @property
    def nested_samples(self):
        return self.results['combined']['nested_samples']

    @nested_samples.setter
    def nested_samples(self, filename=None):
        self.postprocess(self, filename=filename)

    @property
    def posterior_samples(self):
        return self.results['combined']['posterior_samples']

    @posterior_samples.setter
    def posterior_samples(self, filename=None):
        self.postprocess(self, filename=filename)

    @property
    def prior_samples(self):
        return self.results['combined']['prior_samples']

    @prior_samples.setter
    def prior_samples(self, filename=None):
        self.postprocess(self, filename=filename)

    @property
    def logZ(self):
        return self.results['combined']['logZ']

    @logZ.setter
    def logZ(self, filename=None):
        self.postprocess(self, filename=filename)

    @property
    def logZ_error(self):
        return self.results['combined']['logZ_error']

    @logZ_error.setter
    def logZ_error(self, filename=None):
        self.postprocess(self, filename=filename)

    @property
    def information(self):
        return self.results['combined']['information']

    @information.setter
    def information(self, filename=None):
        self.postprocess(self, filename=filename)

    def plot(self, corner = True):
        """
        Make diagnostic plots of the posterior and nested samples
        """
        from . import plot
        for n in self.posterior_samples.dtype.names:
            plot.plot_hist(self.posterior_samples[n].ravel(), name = n,
                           prior_samples = self.prior_samples[n].ravel(),
                           filename = os.path.join(self.output,'posterior_{0}.pdf'.format(n)))

        plot.trace_plot([self.results['run_{}'.format(i)]['nested_samples'] for i in range(self.nnest)],
                        [self.nlive]*self.nnest, self.output)

        if corner:
            import numpy as np
            plotting_posteriors = np.squeeze(self.posterior_samples.view((self.posterior_samples.dtype[0], len(self.posterior_samples.dtype.names))))

            plot.plot_corner(plotting_posteriors,

                             labels=self.prior_samples.dtype.names,
                             filename=os.path.join(self.output,'corner.pdf'))

        live_points = ray.get([self.ns_pool[v].get_live_points.remote() for v in range(self.nnest)])
        
        for i in range(self.nnest):
            plot.plot_indices(live_points[i].get_insertion_indices(), filename=os.path.join(self.output, 'insertion_indices_{}.pdf'.format(i)))

    def checkpoint(self):
        """
        checkpoint the nested samplers
        """
        for i,s in enumerate(self.ns_pool):
            self.logger.critical('Checkpointing nested sampling {}'.format(s))
        ray.get([s.save.remote() for s in self.ns_pool])

    def load_nested_sampler_state(self, resume_file):
        """
        load the nested samplers dictionary state
        """
        self.logger.critical('Loading nested sampling state from {}'.format(resume_file))
        with open(resume_file,"rb") as f:
            obj = dill.load(f)
        return obj



# cython: language_level=3, boundscheck=False, wraparound=False, infer_types=True, embed_signature=True
cimport numpy as np

cdef class LivePoint:
    cdef public double logL
    cdef public double logP
    cdef public unsigned int dimension
    cdef public np.ndarray values
    cdef public list names
    cdef LivePoint _copy(LivePoint self)
    cdef np.ndarray _asnparray(LivePoint self)



import unittest
import numpy as np
import raynest.model
from numpy import log

class AckleyModel(raynest.model.Model):
    """
    Ackley problem from https://en.wikipedia.org/wiki/Test_functions_for_optimization
    """
    names=['x','y']
    bounds=[[-5,5],[-5,5]]
    data = None

    def log_likelihood(self,x):
        return log(ackley(x['x'],x['y']))

    def force(self,x):
        f = np.zeros(1, dtype = {'names':x.names, 'formats':['f8' for _ in x.names]})
        return f

def ackley(x, y):
    """
    Ackley problem
    """
    r = np.sqrt(0.5*(x*x+y*y))
    first = 20.0*np.exp(r)
    second = np.exp(0.5*(np.cos(2.0*np.pi*x)+np.cos(2.0*np.pi*y)))
    return (first+second-np.exp(1)-20)

class AckleyTestCase(unittest.TestCase):
    """
    Test the gaussian model
    """
    def setUp(self):
        self.work=raynest.raynest(AckleyModel(),verbose=1,nensemble=8,nlive=1000,maxmcmc=1000)

    def test_run(self):
        self.work.run()



def test_all():
    unittest.main(verbosity=2)

if __name__=='__main__':
    unittest.main(verbosity=2)



import os
import gc
import logging
import psutil

# Default formats and level names
FORMATTER = logging.Formatter(
    '%(asctime)s - %(name)-38s: %(message)s',
    datefmt='%Y-%m-%d, %H:%M:%S'
)
LEVELS = ['CRITICAL', 'WARNING', 'INFO', 'DEBUG']
LOGGER = logging.getLogger('raynest.utils')


class _Handler(logging.Handler):

    def __init__(self, verbose=0, **kwargs):
        super().__init__(**kwargs)
        self.set_verbosity(verbose)
        self.setFormatter(FORMATTER)

    def get_verbosity(self):
        return self._verbose

    def set_verbosity(self, verbose):
        LOGGER.warning('Setting verbosity to {}'.format(verbose))
        self._verbose = verbose
        self.setLevel(LEVELS[verbose])


class StreamHandler(_Handler, logging.StreamHandler):
    def __init__(self, verbose=0, **kwargs):
        super().__init__(verbose=verbose, **kwargs)


class FileHandler(_Handler, logging.FileHandler):
    def __init__(self, filename, verbose=0, **kwargs):
        super().__init__(filename=filename, verbose=verbose, **kwargs)


class LogFile:
    """
    Context manager for file logging. It logs everything from `logger`
    in some file at a given `filename`.

    Parameters
    ----------
    filename : str
        Filename under which to save the log.

    verbose : int, optional
        Logging level verbosity 0='CRITICAL' 1='WARNING' 2='INFO' 3='DEBUG'.

    loggername : str, optional
        Name of the logger to send to file at `path`. Default is `'raynest'` so
        all raynest logs are recorded. E.g. specify `'raynest.raynest'` to only
        record logs from the `raynest.py` module.

    Attributes
    ----------
    handler : logging.FileHandler
        File handler object.

    Examples
    --------
    ```python
    from raynest.utils import LogFile

    with LogFile('example.log') as flog:
        # Do some stuff here and it will be logged to 'example.log'
        ...

    # Do some stuff here and it won't be logged to 'example.log'

    with flog:
        # Do some stuff here and it will be logged to 'example.log'
        ...
    ```

    """

    def __init__(self, filename, verbose=0, loggername='raynest'):
        self._filename = filename
        self._verbose = verbose
        self._logger = logging.getLogger(loggername)
        self.handler = None

    def open(self):
        self.handler = FileHandler(self._filename, verbose=self._verbose)
        self._logger.addHandler(self.handler)

    def close(self):
        self._logger.removeHandler(self.handler)
        self.handler.close()
        self.handler = None

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        self.close()

def auto_garbage_collect(pct=80.0):
    """
    auto_garbage_collection - Call the garbage collection if memory used is greater than 80% of total available memory.
                              This is called to deal with an issue in Ray not freeing up used memory.

        pct - Default value of 80%.  Amount of memory in use that triggers the garbage collection call.
    """
    if psutil.virtual_memory().percent >= pct:
        gc.collect()
    return



import os
import platform
import numpy
from setuptools import setup, Extension, find_packages
from Cython.Build import cythonize

WINDOWS = platform.system().lower() == "windows"

# set extension
libraries = [] if WINDOWS else ["m"]
ext_modules = [Extension("raynest.parameter",
                         sources=[os.path.join("raynest", "parameter.pyx")],
                         include_dirs=['raynest', numpy.get_include()],
                         libraries=libraries,
                         extra_compile_args=["-O3","-ffast-math","-mavx2","-ftree-vectorize"])]
ext_modules = cythonize(ext_modules)

setup(url="https://github.com/wdpozzo/raynest",  # Replace with your package URL
      packages=find_packages(),
      classifiers=[
                'Development Status :: 4 - Beta',
                'License :: OSI Approved :: MIT License',
                'Programming Language :: Python :: 3.9'
                'Programming Language :: Python :: 3.10'
                ],
      python_requires='>=3.9',
      ext_modules=ext_modules)




from __future__ import print_function
from sys import getsizeof, stderr
from itertools import chain
from collections import deque
try:
    from reprlib import repr
except ImportError:
    pass

def total_size(o, handlers={}, verbose=False):
    """ Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    """
    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {tuple: iter,
                    list: iter,
                    deque: iter,
                    dict: dict_handler,
                    set: iter,
                    frozenset: iter,
                   }
    all_handlers.update(handlers)     # user handlers take precedence
    seen = set()                      # track which object id's have already been seen
    default_size = getsizeof(0)       # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:       # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        else:
            if not hasattr(o.__class__, '__slots__'):
                if hasattr(o, '__dict__'):
                    s+=sizeof(o.__dict__) # no __slots__ *usually* means a __dict__, but some special builtin classes (such as `type(None)`) have neither
                # else, `o` has no attributes at all, so sys.getsizeof() actually returned the correct value
            else:
                s+=sum(sizeof(getattr(o, x)) for x in o.__class__.__slots__ if hasattr(o, x))
        return s

    return sizeof(o)


##### Example call #####

if __name__ == '__main__':
    d = dict(a=1, b=2, c=3, d=[4,5,6,7], e='a string of chars')
    print(total_size(d, verbose=True))



import os
import matplotlib as mpl
from matplotlib import pyplot as plt
import numpy as np

def init_plotting():
    """
    utility function to set up the
    plotting environment
    """
    plt.rcParams['figure.figsize'] = (3.4, 3.4)
    plt.rcParams['font.size'] = 11
    plt.rcParams['font.family'] = 'DejaVu Sans'
    plt.rcParams['font.sans-serif'] = ['Bitstream Vera Sans']
    plt.rcParams['axes.labelsize'] = plt.rcParams['font.size']
    plt.rcParams['axes.titlesize'] = plt.rcParams['font.size']
    plt.rcParams['legend.fontsize'] = 10
    plt.rcParams['xtick.labelsize'] = 9
    plt.rcParams['ytick.labelsize'] = 9
    plt.rcParams['xtick.major.size'] = 3
    plt.rcParams['xtick.minor.size'] = 3
    plt.rcParams['xtick.major.width'] = 1
    plt.rcParams['xtick.minor.width'] = 1
    plt.rcParams['ytick.major.size'] = 3
    plt.rcParams['ytick.minor.size'] = 3
    plt.rcParams['ytick.major.width'] = 1
    plt.rcParams['ytick.minor.width'] = 1
    plt.rcParams['legend.frameon'] = False
    plt.rcParams['legend.loc'] = 'center left'
    plt.rcParams['axes.linewidth'] = 1
    plt.rcParams['contour.negative_linestyle'] = 'solid'
    plt.gca().spines['right'].set_color('none')
    plt.gca().spines['top'].set_color('none')
    plt.gca().xaxis.set_ticks_position('bottom')
    plt.gca().yaxis.set_ticks_position('left')

init_plotting()

def trace_plot(nested_samples, nlives, folder):
    """
    produce trace plots for each of the nested sampling
    parallel runs
    """
    from matplotlib import cm
    rows  = len(nested_samples)
    names = nested_samples[0].dtype.names
    
    for n in names:
        fig, axs = plt.subplots(rows, 1, squeeze=False)

        for i,s in enumerate(nested_samples):
            S = axs[i,0].scatter(1+np.arange(len(s[n])),s[n],c=s['logPrior'],cmap=cm.inferno,s=2)
            try:
                CB = plt.colorbar(S, ax=axs[i,0])
                CB.set_label('logPrior')
            except:
                pass
            axs[i,0].set_xlabel('iteration number')
            axs[i,0].set_ylabel(n)
        plt.savefig(os.path.join(folder,n+"_trace.png"), bbox_inches='tight')
        plt.close(fig)
        
def plot_chain(x, name=None, filename=None):
    """
    Produce a trace plot from a list of
    chains
    """
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    ax.plot(x,',')
    ax.grid()
    ax.set_xlabel('iteration number')
    if name is not None:
        ax.set_ylabel(name)
        if filename is None:
            filename=name+'_chain.png'
    if filename is not None:
        plt.savefig(filename, bbox_inches='tight')
    plt.close(fig)

def plot_hist(x, name=None, prior_samples=None, mcmc_samples=None, filename=None):
    """
    Produce a histogram
    """
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    ax.hist(x, density = True, color='black', linewidth = 1.25,
                histtype='step', bins=len(x)//50, label = 'posterior')
    if prior_samples is not None:
        ax.hist(prior_samples, density = True, color='green', linewidth = 1.0,
                histtype='step', bins=len(x)//50, label = 'prior')
    if mcmc_samples is not None:
        ax.hist(mcmc_samples, density = True, color='red', linewidth = 1.0,
                histtype='step', bins=len(x)//50, label = 'mcmc')
    ax.legend(loc='upper left')
    ax.set_ylabel('probability density')
    if name is not None:
        ax.set_xlabel(name)
        if filename is None:
            filename=name+'_hist.png'
    if filename is not None:
        plt.savefig(filename, bbox_inches='tight')
    plt.close()


def plot_indices(indices, filename=None, max_bins=30):
    """
    Histogram indices for insertion indices tests.

    Parameters
    ----------
    indices : list
        List of insertion indices
    filename : str, optional
        Filename used to saved resulting figure. If not specified figure
        is not saved.
    max_bins : int, optional
        Maximum number of bins in the histogram.
    """
    fig = plt.figure()
    ax  = fig.add_subplot(111)

    ax.hist(indices, density=True, color='tab:blue', linewidth=1.25,
                histtype='step', bins=min(len(indices) // 100, max_bins))
    # Theoretical distribution
    ax.axhline(1, color='black', linewidth=1.25, linestyle=':', label='pdf')

    ax.legend()
    ax.set_xlabel('Insertion indices [0, 1]')

    if filename is not None:
        plt.savefig(filename, bbox_inches='tight')
    plt.close()


def plot_corner(xs, filename=None, **kwargs):
    """
    Produce a corner plot
    """
    import corner
    mask = [i for i in range(xs.shape[-1]) if not all(xs[:,i]==xs[0,i]) ]
    fig = corner.corner(xs[:,mask], color='k', hist_kwargs={'density':True}, **kwargs)


    if filename is not None:
        plt.savefig(filename,bbox_inches='tight')
    plt.close()




import unittest
import numpy as np
import raynest.model

"""
Once run this example, use

pyprof2calltree -i <profile report file> -k

for detail call tree information
"""

from eggbox import EggboxModel

if __name__=='__main__':
    work=raynest.raynest(EggboxModel(),verbose=2,nthreads=4,nlive=1000,maxmcmc=5000)
    work.profile()
    #unittest.main(verbosity=2)
